"""Test Oklab library."""
import unittest
from . import util
from coloraide import Color as Color, NaN
import pytest


class TestsOklab(util.ColorAssertsPyTest):
    """Test Oklab."""

    COLORS = [
        # Test general conversion
        ('red', 'color(--oklab 0.62796 0.22486 0.12585)'),
        ('orange', 'color(--oklab 0.79269 0.05661 0.16138)'),
        ('yellow', 'color(--oklab 0.96798 -0.07137 0.19857)'),
        ('green', 'color(--oklab 0.51975 -0.1403 0.10768)'),
        ('blue', 'color(--oklab 0.45201 -0.03246 -0.31153)'),
        ('indigo', 'color(--oklab 0.33898 0.09416 -0.15255)'),
        ('violet', 'color(--oklab 0.7619 0.15647 -0.1008)'),
        ('white', 'color(--oklab 1 0 0)'),
        ('gray', 'color(--oklab 0.59987 0 0)'),
        ('black', 'color(--oklab 0 0 0)'),
        # Test CSS
        ('oklab(1 0.1 -0.1)', 'color(--oklab 1 0.1 -0.1)'),
        ('oklab(1 0.1 -0.1 / 0.5)', 'color(--oklab 1 0.1 -0.1 / 0.5)'),
        ('oklab(50% 0.2 -0.2)', 'color(--oklab 0.5 0.2 -0.2)'),
        ('oklab(50% 50% -50% / 50%)', 'color(--oklab 0.5 0.2 -0.2 / 0.5)'),
        ('oklab(none none none / none)', 'color(--oklab none none none / none)'),
        ('oklab(1, 0.1, -0.1)', None),
        # Test color
        ('color(--oklab 1 0.1 -0.1)', 'color(--oklab 1 0.1 -0.1)'),
        ('color(--oklab 1 0.1 -0.1 / 0.5)', 'color(--oklab 1 0.1 -0.1 / 0.5)'),
        ('color(--oklab 50% 50% -50% / 50%)', 'color(--oklab 0.5 0.2 -0.2 / 0.5)'),
        ('color(--oklab none none none / none)', 'color(--oklab none none none / none)'),
        # Test percent ranges
        ('color(--oklab 0% 0% 0%)', 'color(--oklab 0 0 0)'),
        ('color(--oklab 100% 100% 100%)', 'color(--oklab 1 0.4 0.4)'),
        ('color(--oklab -100% -100% -100%)', 'color(--oklab -1 -0.4 -0.4)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        if color2 is None:
            with pytest.raises(ValueError):
                Color(color1)
        else:
            self.assertColorEqual(Color(color1).convert('oklab'), Color(color2), color=True)


class TestOklabSerialize(util.ColorAssertsPyTest):
    """Test Oklab serialization."""

    COLORS = [
        # Test hex no options
        ('oklab(0.75 0.1 -0.1)', {}, 'oklab(0.75 0.1 -0.1)'),
        # Test alpha
        ('oklab(0.75 0.1 -0.1 / 0.5)', {}, 'oklab(0.75 0.1 -0.1 / 0.5)'),
        ('oklab(0.75 0.1 -0.1)', {'alpha': True}, 'oklab(0.75 0.1 -0.1 / 1)'),
        ('oklab(0.75 0.1 -0.1 / 0.5)', {'alpha': False}, 'oklab(0.75 0.1 -0.1)'),
        # Test percent
        ('oklab(0.5 0.2 -0.2)', {'percent': True}, 'oklab(50% 50% -50%)'),
        ('oklab(0.5 0.2 -0.2 / 0.5)', {'percent': True, 'alpha': True}, 'oklab(50% 50% -50% / 0.5)'),
        # Test None
        ('oklab(none 0.1 -0.1)', {}, 'oklab(0 0.1 -0.1)'),
        ('oklab(none 0.1 -0.1)', {'none': True}, 'oklab(none 0.1 -0.1)'),
        # Test fit (not bound)
        ('oklab(0.2 0.5 0)', {}, 'oklab(0.2 0.5 0)'),
        ('oklab(0.2 0.5 0)', {'fit': False}, 'oklab(0.2 0.5 0)'),
        # Test color
        ('oklab(none 0.1 -0.1 / 0.5)', {'color': True}, 'color(--oklab 0 0.1 -0.1 / 0.5)'),
        ('oklab(none 0.1 -0.1)', {'color': True, 'none': True}, 'color(--oklab none 0.1 -0.1)'),
        ('oklab(0 0.1 -0.1)', {'color': True, 'alpha': True}, 'color(--oklab 0 0.1 -0.1 / 1)'),
        ('oklab(0 0.1 -0.1 / 0.5)', {'color': True, 'alpha': False}, 'color(--oklab 0 0.1 -0.1)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestOklabProperties(util.ColorAsserts, unittest.TestCase):
    """Test Oklab."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--oklab 1 0.2 -0.3 / 1)')
        self.assertEqual(c['lightness'], 1)
        c['lightness'] = 0.2
        self.assertEqual(c['lightness'], 0.2)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--oklab 1 0.2 -0.3 / 1)')
        self.assertEqual(c['a'], 0.2)
        c['a'] = 0.1
        self.assertEqual(c['a'], 0.1)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--oklab 1 0.2 -0.3 / 1)')
        self.assertEqual(c['b'], -0.3)
        c['b'] = 0.1
        self.assertEqual(c['b'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--oklab 1 0.2 -0.3 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('oklab', [0.3, 0, 0]).is_achromatic(), True)
        self.assertEqual(Color('oklab', [-0.3, 0, 0]).is_achromatic(), True)
        self.assertEqual(Color('oklab', [0.3, 0.000001, 0]).is_achromatic(), True)
        self.assertEqual(Color('oklab', [NaN, 0.00001, 0]).is_achromatic(), True)
        self.assertEqual(Color('oklab', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('oklab', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('oklab', [0, 0.3, -0.4]).is_achromatic(), False)
        self.assertEqual(Color('oklab', [NaN, 0, -0.3]).is_achromatic(), False)
        self.assertEqual(Color('oklab', [0.3, NaN, 0]).is_achromatic(), True)
        self.assertEqual(Color('oklab', [NaN, NaN, 0]).is_achromatic(), True)



"""Test Lab library."""
import unittest
from . import util
from coloraide import Color, NaN
import pytest


class TestLab(util.ColorAssertsPyTest):
    """Test Lab."""

    COLORS = [
        ('red', 'color(--lab 54.291 80.805 69.891)'),
        ('orange', 'color(--lab 75.59 27.516 79.121)'),
        ('yellow', 'color(--lab 97.607 -15.75 93.394)'),
        ('green', 'color(--lab 46.278 -47.552 48.586)'),
        ('blue', 'color(--lab 29.568 68.287 -112.03)'),
        ('indigo', 'color(--lab 19.715 47.029 -54.278)'),
        ('violet', 'color(--lab 69.618 53.295 -36.538)'),
        ('white', 'color(--lab 100 0 0)'),
        ('gray', 'color(--lab 53.585 0 0)'),
        ('black', 'color(--lab 0 0 0)'),
        # Test CSS
        ('lab(100 10 -10)', 'color(--lab 100 10 -10)'),
        ('lab(100 10 -10 / 0.5)', 'color(--lab 100 10 -10 / 0.5)'),
        ('lab(50% 10 -10)', 'color(--lab 50 10 -10)'),
        ('lab(50% 50% -50% / 50%)', 'color(--lab 50 62.5 -62.5 / 0.5)'),
        ('lab(none none none / none)', 'color(--lab none none none / none)'),
        ('lab(1, 10, -10)', None),
        ('lab(1 10)', None),
        ('lab(1 10deg -10)', None),
        ('lab(1 10 -10 50%)', None),
        # Test color
        ('color(--lab 100 10 -10)', 'color(--lab 100 10 -10)'),
        ('color(--lab 100 10 -10 / 0.5)', 'color(--lab 100 10 -10 / 0.5)'),
        ('color(--lab 50% 50% -50% / 50%)', 'color(--lab 50 62.5 -62.5 / 0.5)'),
        ('color(--lab none none none / none)', 'color(--lab none none none / none)'),
        # Test percent ranges
        ('color(--lab 0% 0% 0%)', 'color(--lab 0 0 0)'),
        ('color(--lab 100% 100% 100%)', 'color(--lab 100 125 125)'),
        ('color(--lab -100% -100% -100%)', 'color(--lab -100 -125 -125)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        if color2 is None:
            with pytest.raises(ValueError):
                Color(color1)
        else:
            self.assertColorEqual(Color(color1).convert('lab'), Color(color2), color=True)


class TestLabSerialize(util.ColorAssertsPyTest):
    """Test Lab serialization."""

    COLORS = [
        # Test hex no options
        ('lab(75 10 -10)', {}, 'lab(75 10 -10)'),
        # Test alpha
        ('lab(75 10 -10 / 0.5)', {}, 'lab(75 10 -10 / 0.5)'),
        ('lab(75 10 -10)', {'alpha': True}, 'lab(75 10 -10 / 1)'),
        ('lab(75 10 -10 / 0.5)', {'alpha': False}, 'lab(75 10 -10)'),
        # Test percent
        ('lab(50 62.5 -62.5)', {'percent': True}, 'lab(50% 50% -50%)'),
        ('lab(50 62.5 -62.5 / 0.5)', {'percent': True, 'alpha': True}, 'lab(50% 50% -50% / 0.5)'),
        # Test None
        ('lab(none 10 -10)', {}, 'lab(0 10 -10)'),
        ('lab(none 10 -10)', {'none': True}, 'lab(none 10 -10)'),
        # Test fit (not bound)
        ('lab(20 130 0)', {}, 'lab(20 130 0)'),
        ('lab(20 130 0)', {'fit': False}, 'lab(20 130 0)'),
        # Test color
        ('lab(none 10 -10 / 0.5)', {'color': True}, 'color(--lab 0 10 -10 / 0.5)'),
        ('lab(none 10 -10)', {'color': True, 'none': True}, 'color(--lab none 10 -10)'),
        ('lab(0 10 -10)', {'color': True, 'alpha': True}, 'color(--lab 0 10 -10 / 1)'),
        ('lab(0 10 -10 / 0.5)', {'color': True, 'alpha': False}, 'color(--lab 0 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestLabProperties(util.ColorAsserts, unittest.TestCase):
    """Test Lab."""

    def test_names(self):
        """Test LCh-ish names."""

        self.assertEqual(Color('color(--lab 90% 50 -20 / 1)')._space.names(), ('l', 'a', 'b'))

    def test_indexes(self):
        """Test LCh-ish names."""

        self.assertEqual(Color('color(--lab 90% 50 -20 / 1)')._space.indexes(), [0, 1, 2])

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--lab 90% 50 -20 / 1)')
        self.assertEqual(c['lightness'], 90)
        c['lightness'] = 80
        self.assertEqual(c['lightness'], 80)

    def test_a(self):
        """Test `a`."""

        c = Color('color(--lab 90% 50 -20 / 1)')
        self.assertEqual(c['a'], 50)
        c['a'] = 40
        self.assertEqual(c['a'], 40)

    def test_b(self):
        """Test `b`."""

        c = Color('color(--lab 90% 50 -20 / 1)')
        self.assertEqual(c['b'], -20)
        c['b'] = -10
        self.assertEqual(c['b'], -10)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--lab 90% 50 -20 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('lab', [30, 0, 0]).is_achromatic(), True)
        self.assertEqual(Color('lab', [30, 0.000001, 0]).is_achromatic(), True)
        self.assertEqual(Color('lab', [NaN, 0.00001, 0]).is_achromatic(), True)
        self.assertEqual(Color('lab', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('lab', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('lab', [0, 30, -40]).is_achromatic(), False)
        self.assertEqual(Color('lab', [NaN, 0, -30]).is_achromatic(), False)
        self.assertEqual(Color('lab', [30, NaN, 0]).is_achromatic(), True)
        self.assertEqual(Color('lab', [NaN, NaN, 0]).is_achromatic(), True)



"""
WCAG 2.1 contrast ratio.

https://www.w3.org/TR/WCAG20/#contrast-ratiodef
"""
from __future__ import annotations
from ..contrast import ColorContrast
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class WCAG21Contrast(ColorContrast):
    """WCAG 2.1 contrast ratio."""

    NAME = "wcag21"

    def contrast(self, color1: Color, color2: Color, **kwargs: Any) -> float:
        """Contrast."""

        lum1 = max(0, color1.luminance())
        lum2 = max(0, color2.luminance())
        if (lum1 > lum2):
            lum1, lum2 = lum2, lum1
        return (lum2 + 0.05) / (lum1 + 0.05)



"""Piecewise linear interpolation."""
from __future__ import annotations
from .. import algebra as alg
from ..interpolate import Interpolator, Interpolate
from ..types import Vector
from typing import Callable, Mapping, Any, Sequence, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class InterpolatorLinear(Interpolator):
    """Interpolate multiple ranges of colors using linear, Piecewise interpolation."""

    def setup(self) -> None:
        """Setup for linear interpolation."""

        end = self.length - 2

        i = 0
        while i <= end:
            c1, c2 = self.coordinates[i:i + 2]

            # Piecewise interpolation will evaluate the same color two different ways
            # if sandwiched between two other colors. Create pairs and evaluate
            # the undefined values and premultiplied states with this context.
            if i < end:
                self.coordinates.insert(i + 2, c2[:])
                end += 1

            # If we have a NaN for one alpha and the other alpha is not
            # Use the non-NaN alpha, but if we are premultiplied, we need
            # to now premultiply that coordinate set.
            if self.premultiplied:
                a, b = c1[-1], c2[-1]
                a_nan, b_nan = alg.is_nan(a), alg.is_nan(b)

                # Premultiply the alpha
                if not a_nan:
                    self.premultiply(c1)

                    # Mirror the alpha of its sibling as it has no alpha
                    if b_nan:
                        self.premultiply(c2, a)

                # Premultiply the alpha
                if not b_nan:
                    self.premultiply(c2)

                    # Mirror the alpha of its sibling as it has no alpha
                    if a_nan:
                        self.premultiply(c1, b)

            i += 2

    def interpolate(
        self,
        point: float,
        index: int
    ) -> Vector:
        """Interpolate."""

        # Interpolate between the values of the two colors for each channel.
        channels = []
        i = (index - 1) * 2

        for i, values in enumerate(zip(*self.coordinates[i:i + 2])):  # noqa: B020
            a, b = values

            # Both values are undefined, so return undefined
            if alg.is_nan(a) and alg.is_nan(b):
                value = alg.NaN

            # One channel is undefined, take the one that is not
            elif alg.is_nan(a):
                value = b
            elif alg.is_nan(b):
                value = a

            # Using linear interpolation between the two points
            else:
                # Interpolate
                value = alg.lerp(a, b, self.ease(point, i))
            channels.append(value)

        return channels


class Linear(Interpolate):
    """Linear interpolation plugin."""

    NAME = "linear"

    def interpolator(
        self,
        coordinates: list[Vector],
        channel_names: Sequence[str],
        create: type[Color],
        easings: list[Callable[..., float] | None],
        stops: dict[int, float],
        space: str,
        out_space: str,
        progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
        premultiplied: bool,
        extrapolate: bool = False,
        domain: list[float] | None = None,
        padding: float | tuple[float, float] | None = None,
        **kwargs: Any
    ) -> Interpolator:
        """Return the linear interpolator."""

        return InterpolatorLinear(
            coordinates,
            channel_names,
            create,
            easings,
            stops,
            space,
            out_space,
            progress,
            premultiplied,
            extrapolate,
            domain,
            padding,
            **kwargs
        )



"""
ICtCp class.

https://professional.dolby.com/siteassets/pdfs/ictcp_dolbywhitepaper_v071.pdf
"""
from __future__ import annotations
from .lab import Lab
from ..cat import WHITES
from ..channels import Channel, FLG_MIRROR_PERCENT
from .. import util
from .. import algebra as alg
from ..types import Vector

# All PQ Values are equivalent to defaults as stated in link below:
# https://en.wikipedia.org/wiki/High-dynamic-range_video#Perceptual_quantizer
#
# ```
# M1 = 2610 / 16384
# M1INV = 16384 / 2610
# M2 = 2523 / 32
# M2INV = 32 / 2523
# C1 = 3424 / 4096
# C2 = 2413 / 128
# C3 = 2392 / 128
# ```

# XYZ transform matrices
xyz_to_lms_m = [
    [0.359132, 0.697604, -0.03578],
    [-0.19218800000000003, 1.1003800000000001, 0.07554],
    [0.006956, 0.074916, 0.8433400000000001]
]

lms_to_xyz_mi = [
    [2.070508203420414, -1.3267039449989098, 0.20668057903526463],
    [0.36502513723373864, 0.6804585253538307, -0.045463558701123154],
    [-0.04950397021841152, -0.049503970218411505, 1.1880952852418765]
]

# LMS to Izazbz matrices
lms_p_to_ictcp_m = [
    [0.5, 0.5, 0.0],
    [1.61376953125, -3.323486328125, 1.709716796875],
    [4.378173828125, -4.24560546875, -0.132568359375]
]

ictcp_to_lms_p_mi = [
    [1.0, 0.008609037037932726, 0.11102962500302596],
    [1.0, -0.008609037037932726, -0.11102962500302596],
    [1.0000000000000002, 0.5600313357106792, -0.32062717498731885]
]


def ictcp_to_xyz_d65(ictcp: Vector) -> Vector:
    """From ICtCp to XYZ."""

    # Convert to LMS prime
    pqlms = alg.dot(ictcp_to_lms_p_mi, ictcp, dims=alg.D2_D1)

    # Decode PQ LMS to LMS
    lms = util.pq_st2084_eotf(pqlms)

    # Convert back to absolute XYZ D65
    absxyz = alg.dot(lms_to_xyz_mi, lms, dims=alg.D2_D1)

    # Convert back to normal XYZ D65
    return util.absxyz_to_xyz(absxyz)


def xyz_d65_to_ictcp(xyzd65: Vector) -> Vector:
    """From XYZ to ICtCp."""

    # Convert from XYZ D65 to an absolute XYZ D5
    absxyz = util.xyz_to_absxyz(xyzd65)

    # Convert to LMS
    lms = alg.dot(xyz_to_lms_m, absxyz, dims=alg.D2_D1)

    # PQ encode the LMS
    pqlms = util.pq_st2084_oetf(lms)

    # Calculate Izazbz
    return alg.dot(lms_p_to_ictcp_m, pqlms, dims=alg.D2_D1)


class ICtCp(Lab):
    """ICtCp class."""

    BASE = "xyz-d65"
    NAME = "ictcp"
    SERIALIZE = ("--ictcp",)
    CHANNELS = (
        Channel("i", 0.0, 1.0),
        Channel("ct", -0.5, 0.5, flags=FLG_MIRROR_PERCENT),
        Channel("cp", -0.5, 0.5, flags=FLG_MIRROR_PERCENT)
    )
    CHANNEL_ALIASES = {
        "intensity": "i",
        "protan": "cp",
        "tritan": "ct"
    }
    WHITE = WHITES['2deg']['D65']
    DYNAMIC_RANGE = 'hdr'

    def names(self) -> tuple[str, ...]:
        """
        Return Lab-ish names in the order L a b.

        ICtCp flips protan and tritan (which corresponds to `a` and `b`).
        Interestingly, the polarity of tritan (yellow/blue) is also flipped.
        """

        channels = self.channels
        return channels[0], channels[2], channels[1]

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from ICtCp."""

        return ictcp_to_xyz_d65(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to ICtCp."""

        return xyz_d65_to_ictcp(coords)



"""
HCT color space.

This implements the HCT color space as described. This is not a port of the Material library.
We simply, as described, create a color space with CIELAB L* and CAM16's C and h components.
Environment settings are calculated with the assumption of L* 50.

Our approach getting back to sRGB is likely different as we are using a simple binary search
to help calculate the missing piece (J) needed to convert CAM16 C and h back to XYZ such that
Y matches the L* conversion back to Y. I am certain the Material library is most likely
employing further tricks to resolve faster. We also try to get as good round trip as possible,
which forces us to probably make more iterations than Material does (none of this has been confirmed).

As ColorAide usually cares about setting powerless hues as NaN, especially for good interpolation,
we've also calculated the cut off for chromatic colors and will properly enforce achromatic,powerless
hues. This is because CAM16 actually resolves colors as achromatic before chroma reaches zero as
lightness increases. In the SDR range, a Tone of 100 will have a cut off as high as ~2.87 chroma.

Generally, the HCT color space is restricted to SDR range in the Material library, but we do not have
such restrictions.

Though we did not port HCT from Material Color Utilities, we did test against it, and are pretty
much on point. The only differences are due to matrix precision and white point precision. Material
uses an RGB <-> XYZ matrix that rounds values off significantly more than we do. Also, while we
calculate the XYZ points from the `xy` points without rounding, they have rounded XYZ points. Lastly,
the gamut mapping algorithm we use is likely different even though it arrives at pretty much the same
result, so slightly different values can occur.

Material:

```
> hct.Hct.fromInt(0xff305077)
Hct {
  argb: 4281356407,
  internalHue: 256.8040416857594,
  internalChroma: 31.761442797741243,
  internalTone: 33.34501410942328
}
```

ColorAide:

```
>>> from coloraide_extras.everything import ColorAll as Color
>>> Color('#305077').convert('hct')
color(--hct 256.79 31.766 33.344 / 1)
```

Differences are inconsequential.
"""
from __future__ import annotations
from .. import algebra as alg
from .. import util
from ..spaces import Space, LChish
from ..cat import WHITES
from ..channels import Channel, FLG_ANGLE
from .cam16_jmh import Environment, cam16_to_xyz_d65, xyz_d65_to_cam16
from .cam16_jmh import Achromatic as _Achromatic
from .srgb_linear import lin_srgb_to_xyz
from .srgb import lin_srgb
from .lab import EPSILON, KAPPA, KE
from ..types import Vector, VectorLike
from typing import Any
import math

ACHROMATIC_RESPONSE = [
    [0.06991457401674239, 0.14894004649491988, 209.54685746644427],
    [0.13982914803348123, 0.19600973408108388, 209.54683151374996],
    [0.20974372205022362, 0.22847588408235084, 209.54681244250116],
    [0.279658296066966, 0.2539510473857674, 209.54679680373425],
    [0.34957287008370486, 0.27520353123074354, 209.5467833054029],
    [0.6991457401674133, 0.3503388633000292, 209.54673233806605],
    [1.0487186102511181, 0.40138459372769575, 209.54669489478562],
    [1.398291480334823, 0.4411483978448841, 209.5466641968681],
    [1.7478643504185314, 0.4741765535239112, 209.54663770502128],
    [2.0974372205022362, 0.5026634742613167, 209.54661414586045],
    [2.447010090585941, 0.527852488793119, 209.5465927758291],
    [2.7965829606696495, 0.5505226723841257, 209.54657311722175],
    [3.1624547958278697, 0.5721193279495612, 209.54655401960414],
    [3.5553195764898433, 0.5933528830664249, 209.5465348955287],
    [3.9752712774319967, 0.6142212187429535, 209.5465157686296],
    [4.422818762249047, 0.6347462660662626, 209.54649663922487],
    [4.89845719045681, 0.6549477974515944, 209.5464775076281],
    [5.40266895987223, 0.6748437111130419, 209.54645837410766],
    [5.93592454797767, 0.6944502696114526, 209.54643923893252],
    [6.4986832666748775, 0.7137823011126716, 209.5464201023271],
    [7.091393942308237, 0.7328533701575941, 209.54640096452033],
    [7.714495530830426, 0.7516759233486977, 209.54638182570906],
    [8.362902589291522, 0.7702614142702398, 209.5463626860871],
    [9.010442756551821, 0.7886204111202818, 209.54634354582433],
    [9.653817900540862, 0.8067626898669608, 209.5463244050959],
    [10.29318377392433, 0.8246973152252656, 209.5463052640459],
    [10.928685772528809, 0.8424327113314979, 209.5462861228275],
    [11.56045990779819, 0.8599767236656302, 209.54626698157904],
    [12.188633662809302, 0.8773366735026968, 209.5462478404268],
    [12.813326748632115, 0.8945194059620538, 209.54622869949029],
    [13.434651775012771, 0.9115313325460387, 209.54620955888788],
    [14.0527148470809, 0.928378468919685, 209.54619041872456],
    [14.667616097925809, 0.9450664685622908, 209.5461712791161],
    [15.279450165363095, 0.9616006528304665, 209.54615214014538],
    [15.888306619956744, 0.9779860378836446, 209.5461330019187],
    [16.494270350320917, 0.9942273588675171, 209.54611386451774],
    [17.097421910858294, 1.0103290916839158, 209.54609472803028],
    [17.697837836366574, 1.0262954726377385, 209.5460755925407],
    [18.295590927334914, 1.042130516206454, 209.54605645811773],
    [18.890750509238096, 1.0578380311468378, 209.54603732484748],
    [19.483382668700443, 1.0734216351263122, 209.54601819278727],
    [20.073550469030984, 1.088884768038463, 209.54599906201673],
    [20.661314147315657, 1.1042307041468111, 209.54597993259077],
    [53.38896474111432, 1.8932913045660376, 209.54481718319346],
    [100.0, 2.871588955286716, 209.5429359788321],
    [142.21233355267947, 3.6598615904431204, 209.54109066792404],
    [181.74498695762145, 4.33303753284783, 209.53928262364892],
    [219.379559142843, 4.924275900757143, 209.5375118768831],
    [255.55949550426857, 5.452264905961894, 209.53577795810992],
    [290.56862469201246, 5.929002120237057, 209.53408016967768],
    [324.6031878228979, 6.362852583853973, 209.5324177017366],
    [357.8063946931686, 6.7599912702266085, 209.53078968985596],
    [390.2870215828255, 7.125170758147813, 209.52919524657008],
    [422.13028305191517, 7.462166817890066, 209.52763347977915]]  # type: list[Vector]


def y_to_lstar(y: float, white: VectorLike) -> float:
    """Convert XYZ Y to Lab L*."""

    y = y / white[1]
    fy = alg.nth_root(y, 3) if y > EPSILON else (KAPPA * y + 16) / 116
    return (116.0 * fy) - 16.0


def lstar_to_y(lstar: float, white: VectorLike) -> float:
    """Convert Lab L* to XYZ Y."""

    fy = (lstar + 16) / 116
    y = fy ** 3 if lstar > KE else lstar / KAPPA
    return y * white[1]


def hct_to_xyz(coords: Vector, env: Environment) -> Vector:
    """
    Convert HCT to XYZ.

    Use Newton Raphson method to try and converge as quick as possible or
    converge as close as we can. If we don't converge in about 7 iterations,
    we will instead correct the Y in XYZ and re-calculate the J. This will
    incrementally get our J closer. If we do not converge, we will do a final
    round with Newton Raphson one last time with a more accurate J.

    If, for whatever reason, we cannot achieve the accuracy we seek in the
    allotted iterations, just return the closest we were able to get.
    """

    # Threshold of how close is close enough
    threshold = 2e-8

    h, c, t = coords[:]

    # Shortcut out for black
    if t == 0:
        return [0.0, 0.0, 0.0]

    # Calculate the Y we need to target
    y = lstar_to_y(t, env.ref_white)

    # Try to start with a reasonable initial guess for J
    if c < 142:
        # Calculated by curve fitting J vs T. Works well with colors within a mid-sized gamut, but not ultra wide.
        j = 0.00462403 * t ** 2 + 0.51460278 * t + 2.62845677
    else:
        # For ultra wide gamuts we can get a better J by correcting Y in XYZ and then calculating our J
        xyz = cam16_to_xyz_d65(J=t, C=c, h=h, env=env)
        xyz[1] = y
        j = xyz_d65_to_cam16(xyz, env)[0]

    # Try to find a J such that the returned y matches the returned y of the L*
    attempt = 0
    last = alg.INF
    best = j
    while attempt < 16:
        xyz = cam16_to_xyz_d65(J=j, C=c, h=h, env=env)

        # If we are within range, return XYZ
        # If we are closer than last time, save the values
        delta = abs(xyz[1] - y)
        if delta < last:
            if delta <= threshold:
                return xyz
            best = j
            last = delta

        # Use Newton Raphson method to see if we can quickly converge (or get as close as we can)
        if (attempt < 7 or attempt >= 13) and xyz[1] != 0:
            # ```
            # f(j_root) = (j ** (1 / 2)) * 0.1
            # f(j) = ((f(j_root) * 100) ** 2) / j - 1 = 0
            # f(j_root) = Y = y / 100
            # f(j) = (y ** 2) / j - 1
            # f'(j) = (2 * y) / j
            # ```
            j = j - (xyz[1] - y) * j / (2 * xyz[1])

        # Correct the lightness in XYZ and then re-calculate J
        else:
            xyz[1] = y
            j = xyz_d65_to_cam16(xyz, env)[0]

        attempt += 1

    # We could not acquire the precision we desired, return our closest attempt.
    return cam16_to_xyz_d65(J=best, C=c, h=h, env=env)  # pragma: no cover


def xyz_to_hct(coords: Vector, env: Environment) -> Vector:
    """Convert XYZ to HCT."""

    t = y_to_lstar(coords[1], env.ref_white)
    c, h = xyz_d65_to_cam16(coords, env)[1:3]
    return [h, max(0.0, c), max(0.0, t)]


class Achromatic(_Achromatic):
    """Test HCT achromatic response."""

    # Lightness and chroma (equivalent) index.
    L_IDX = 2
    C_IDX = 1
    H_IDX = 0

    def convert(self, coords: Vector, *, env: Environment, **kwargs: Any) -> Vector:  # type: ignore[override]
        """Convert to the target color space."""

        return xyz_to_hct(lin_srgb_to_xyz(lin_srgb(coords)), env)


class HCT(LChish, Space):
    """HCT class."""

    BASE = "xyz-d65"
    NAME = "hct"
    SERIALIZE = ("--hct",)
    WHITE = WHITES['2deg']['D65']
    ENV = Environment(
        WHITE,
        200 / math.pi * lstar_to_y(50.0, util.xy_to_xyz(WHITE)),
        lstar_to_y(50.0, util.xy_to_xyz(WHITE)) * 100,
        'average',
        False
    )
    CHANNEL_ALIASES = {
        "lightness": "t",
        "tone": "t",
        "chroma": "c",
        "hue": "h"
    }

    # Achromatic detection
    # Precalculated from:
    # [
    #     (1, 5, 1, 1000.0),
    #     (1, 40, 1, 200.0),
    #     (50, 551, 50, 100.0)
    # ]
    ACHROMATIC = Achromatic(
        ACHROMATIC_RESPONSE,
        0.0097,
        0.0787,
        8.1,
        'catrom',
        env=ENV,
    )

    CHANNELS = (
        Channel("h", 0.0, 360.0, flags=FLG_ANGLE),
        Channel("c", 0.0, 145.0, limit=(0.0, None)),
        Channel("t", 0.0, 100.0, limit=(0.0, None))
    )

    def resolve_channel(self, index: int, coords: Vector) -> float:
        """Resolve channels."""

        if index == 0:
            h = coords[0]
            return self.ACHROMATIC.get_ideal_hue(coords[2]) if math.isnan(h) else h

        elif index == 1:
            c = coords[1]
            return self.ACHROMATIC.get_ideal_chroma(coords[0]) if math.isnan(c) else c

        value = coords[index]
        return self.channels[index].nans if math.isnan(value) else value

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return coords[2] == 0.0 or self.ACHROMATIC.test(coords[2], coords[1], coords[0])

    def names(self) -> tuple[str, ...]:
        """Return LCh-ish names in the order L C h."""

        channels = self.channels
        return channels[2], channels[1], channels[0]

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from CAM16."""

        return hct_to_xyz(coords, self.ENV)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to CAM16."""

        return xyz_to_hct(coords, self.ENV)



"""Test harmonies."""
from coloraide import Color
from . import util
import pytest
from itertools import zip_longest as zipl
import unittest


class TestHarmonies(util.ColorAssertsPyTest):
    """Test harmonies."""

    @pytest.mark.parametrize(
        'color,name,colors,space',
        [
            (
                'red',
                'mono',
                ['rgb(128.04 0 0)',
                 'rgb(189.52 0 0)',
                 'rgb(255 0 0)',
                 'rgb(262.16 94.255 76.67)',
                 'rgb(266.2 140.53 123.15)'],
                None
            ),
            (
                'white',
                'mono',
                ['rgb(128.04 128.04 128.04)',
                 'rgb(158.23 158.23 158.23)',
                 'rgb(189.52 189.52 189.52)',
                 'rgb(221.8 221.8 221.8)',
                 'rgb(255 255 255)'],
                None
            ),
            (
                'black',
                'mono',
                ['rgb(0 0 0)',
                 'rgb(3.2946 3.2946 3.2946)',
                 'rgb(21.957 21.957 21.957)',
                 'rgb(45.705 45.705 45.705)',
                 'rgb(71.554 71.554 71.554)'],
                None
            ),
            (
                'lightyellow',
                'mono',
                ['rgb(148.04 148.04 129.36)',
                 'rgb(182.47 182.47 159.83)',
                 'rgb(218.17 218.17 191.41)',
                 'rgb(255 255 224)',
                 'rgb(254.96 255.07 239.68)'],
                None
            ),
            (
                '#000011',
                'mono',
                ['rgb(0 0 2.3084)',
                 'rgb(0 0 17)',
                 'rgb(6.8796 16.788 39.776)',
                 'rgb(31.005 41.641 63.598)',
                 'rgb(58.097 68.188 88.434)'],
                None
            ),
            (
                'red',
                'complement',
                ['rgb(255 0 0)',
                 'rgb(-144.41 169.17 218.81)'],
                None
            ),
            (
                'red',
                'split',
                ['rgb(255 0 0)',
                 'rgb(-108.26 144.11 271.69)',
                 'rgb(-142.25 179.87 140.26)'],
                None
            ),
            (
                'red',
                'analogous',
                ['rgb(255 0 0)',
                 'rgb(239.47 70.306 -85.526)',
                 'rgb(245.33 -9.3968 132.34)'],
                None
            ),
            (
                'red',
                'triad',
                ['rgb(255 0 0)',
                 'rgb(-90.834 173.97 -5.7497)',
                 'rgb(78.547 111.49 288.59)'],
                None
            ),
            (
                'red',
                'square',
                ['rgb(255 0 0)',
                 'rgb(122.96 152.52 -89.834)',
                 'rgb(-144.41 169.17 218.81)',
                 'rgb(162.78 79.238 265.73)'],
                None
            ),
            (
                'red',
                'rectangle',
                ['rgb(255 0 0)',
                 'rgb(239.47 70.306 -85.526)',
                 'rgb(-144.41 169.17 218.81)',
                 'rgb(-108.26 144.11 271.69)'],
                None
            ),
            (
                'red',
                'complement',
                ['rgb(255 0 0)',
                 'rgb(0 255 255)'],
                'hsl'
            )
        ]
    )
    def test_harmonies(self, color, name, colors, space):
        """Test color harmonies."""

        results = Color(color).harmony(name, space=space, out_space='srgb')
        for c1, c2 in zipl(results, [Color(c) for c in colors]):
            self.assertColorEqual(c1, c2)


class TestHarmonyError(util.ColorAsserts, unittest.TestCase):
    """Test harmony errors."""

    def test_bad_space(self):
        """Test using a harmony with sRGB."""

        with self.assertRaises(ValueError):
            Color('red').harmony('complement', space='srgb')

        with self.assertRaises(ValueError):
            Color('red').harmony('split', space='srgb')

        with self.assertRaises(ValueError):
            Color('red').harmony('analogous', space='srgb')

        with self.assertRaises(ValueError):
            Color('red').harmony('triad', space='srgb')

        with self.assertRaises(ValueError):
            Color('red').harmony('square', space='srgb')

        with self.assertRaises(ValueError):
            Color('red').harmony('rectangle', space='srgb')

        with self.assertRaises(ValueError):
            Color('red').harmony('mono', space='srgb')

    def test_bad_harmony(self):
        """Test bad harmony."""

        with self.assertRaises(ValueError):
            Color('red').harmony('bad')



# -*- coding: utf-8 -*-
"""Calculate the required matrices for simulating color vision deficiency."""
import sys
import os

sys.path.insert(0, os.getcwd())

from coloraide import algebra as alg  # noqa: E402
from coloraide import Color  # noqa: E402

# Smith & Pokorny (1975) 2-deg cone fundamentals
M = [
    [0.15514, 0.54312, -0.03286],
    [-0.15514, 0.45684, 0.03286],
    [0, 0, 0.01608]
]

# Linear sRGB to Judd-Vos corrected XYZ (approximation)
# https://vision.psychol.cam.ac.uk/jdmollon/papers/colourmaps.pdf
RGB_TO_XYZ = [
    [40.9568, 35.5041, 17.9167],
    [21.3389, 70.6743, 7.9868],
    [1.86297, 11.462, 91.2367]
]


def pprint(value):
    """Print the matrix."""

    print('[', end='')
    first = True
    for v in value:
        if first:
            first = False
        else:
            print(',\n ', end='')
        print(v, end='')
    print(']')


####################
# General matrices
####################
T = alg.dot(M, alg.divide(RGB_TO_XYZ, 100))
INV_T = alg.inv(T)

print('===== LRGB to LMS =====')
pprint(T)

print('===== LMS to LRGB =====')
pprint(INV_T)

print('===== XYZ to LMS =====')
pprint(M)

print('===== LMS to XYZ =====')
pprint(alg.inv(M))

####################
# Viénot matrices
# http://vision.psychol.cam.ac.uk/jdmollon/papers/colourmaps.pdf
####################
print('===== VIENOT PROTAN =====')
white = Color('white').convert('srgb-linear')
blue = Color('blue').convert('srgb-linear')
lw, mw, sw = alg.dot(T, white[:-1])
lb, mb, sb = alg.dot(T, blue[:-1])
q1 = (lb * sw - lw * sb) / (mb * sw - mw * sb)
q2 = (lb * mw - lw * mb) / (sb * mw - sw * mb)
sp = [[0.0, q1, q2], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
m = alg.multi_dot([INV_T, sp, T])
pprint(m)

print('===== VIENOT DEUTAN =====')
white = Color('white').convert('srgb-linear')
blue = Color('blue').convert('srgb-linear')
lw, mw, sw = alg.dot(T, white[:-1])
lb, mb, sb = alg.dot(T, blue[:-1])
q1 = (mb * sw - mw * sb) / (lb * sw - lw * sb)
q2 = (mb * lw - mw * lb) / (sb * lw - sw * lb)
sd = [[1.0, 0.0, 0.0], [q1, 0.0, q2], [0.0, 0.0, 1.0]]
m = alg.multi_dot([INV_T, sd, T])
pprint(m)

print('===== VIENOT TRITAN =====')
white = Color('white').convert('srgb-linear')
red = Color('red').convert('srgb-linear')
lw, mw, sw = alg.dot(T, white[:-1])
lb, mb, sb = alg.dot(T, red[:-1])
q1 = (sb * mw - sw * mb) / (lb * mw - lw * mb)
q2 = (sb * lw - sw * lb) / (mb * lw - mw * lb)
st = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [q1, q2, 0.0]]
m = alg.multi_dot([INV_T, st, T])
pprint(m)

####################
# Brettel matrices
# https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.496.7153&rep=rep1&type=pdf
####################
LMS_WHITE = alg.dot(T, [1, 1, 1])

# http://cvrl.ioo.ucl.ac.uk/: CIE 1931 2-deg, XYZ CMFs modified by Judd (1951) and Vos (1978)
LMS_475 = alg.dot(M, [1.3287E-01, 1.1284E-01, 9.422E-01])
LMS_575 = alg.dot(M, [8.4394E-01, 9.1558E-01, 1.9706E-03])
LMS_485 = alg.dot(M, [5.6985E-02, 1.6987E-01, 5.864E-01])
LMS_660 = alg.dot(M, [1.6161E-01, 6.1E-02, 1.1906E-05])

AXIS = alg.identity(3)
NAMES = ['PROTAN', 'DEUTAN', 'TRITAN']
WINGS = [(LMS_475, LMS_575), (LMS_475, LMS_575), (LMS_485, LMS_660)]

for axis, name, wings in zip(AXIS, NAMES, WINGS):
    wing1, wing2 = wings

    v1 = alg.cross(LMS_WHITE, wing1)
    v2 = alg.cross(LMS_WHITE, wing2)
    va = alg.cross(LMS_WHITE, axis)

    # Make sure the wings are on the correct sides
    if alg.dot(va, v1) < 0:
        v1, v2 = v2, v1

    index = axis.index(1.0)
    m1 = alg.identity(3)
    m2 = alg.identity(3)
    if index == 0:
        # Protan
        m1[index] = [0.0, -v1[1] / v1[0], -v1[2] / v1[0]]
        m2[index] = [0.0, -v2[1] / v2[0], -v2[2] / v2[0]]
    elif index == 1:
        # Deutan
        m1[index] = [-v1[0] / v1[1], 0.0, -v1[2] / v1[1]]
        m2[index] = [-v2[0] / v2[1], 0.0, -v2[2] / v2[1]]
    else:
        # Tritan
        m1[index] = [-v1[0] / v1[2], -v1[1] / v1[2], 0.0]
        m2[index] = [-v2[0] / v2[2], -v2[1] / v2[2], 0.0]

    print('===== BRETTEL {} ====='.format(name))
    print('--- Wing 1 (LMS to Linear sRGB included) ---')
    pprint(alg.dot(INV_T, m1))
    print('--- Wing 2 (LMS to Linear sRGB included) ---')
    pprint(alg.dot(INV_T, m2))
    print('--- Separating axis ---')
    print(va)



"""Utilities."""
from __future__ import annotations
import math
from . import algebra as alg
from .types import Vector, VectorLike

DEF_PREC = 5
DEF_FIT_TOLERANCE = 0.000075
DEF_ALPHA = 1.0
DEF_MIX = 0.5
DEF_HUE_ADJ = "shorter"
DEF_INTERPOLATE = "oklab"
DEF_FIT = "lch-chroma"
DEF_HARMONY = "oklch"
DEF_DELTA_E = "76"
DEF_AVERAGE = 'srgb-linear'
DEF_CHROMATIC_ADAPTATION = "bradford"
DEF_CONTRAST = "wcag21"
DEF_CCT = "robertson-1968"

# Maximum luminance in PQ is 10,000 cd/m^2
# Relative XYZ has Y=1 for media white
# BT.2048 says media white Y=203 at PQ 58
#
# This is confirmed here: https://www.itu.int/dms_pub/itu-r/opb/rep/R-REP-BT.2408-3-2019-PDF-E.pdf
YW = 203

# PQ Constants
# https://en.wikipedia.org/wiki/High-dynamic-range_video#Perceptual_quantizer
M1 = 2610 / 16384
M2 = 2523 / 32
C1 = 3424 / 4096
C2 = 2413 / 128
C3 = 2392 / 128


def xy_to_xyz(xy: VectorLike, Y: float = 1.0, scale: float = 1.0) -> Vector:
    """
    Convert `xyY` to `xyz`.

    In many cases, we are dealing with chromaticity values with no Y value,
    in this case, assume 1 unless otherwise specified. Generally, scale is
    also assumed to be between 0 - 1, but allow changing scale if we are
    dealing with things like 0 - 100, etc.
    """

    x, y = xy
    return [0.0, 0.0, 0.0] if y == 0 else [(x * Y) / y, Y, (scale - x - y) * Y / y]


def xyz_to_xyY(xyz: VectorLike, white: VectorLike = (0.0, 0.0)) -> Vector:
    """
    XYZ to `xyY`.

    If a white point chromaticity pair is given, black will be aligned with the achromatic axis.
    """

    x, y, z = xyz
    d = x + y + z
    return [white[0], white[1], y] if d == 0 else [x / d, y / d, y]


def xy_to_uv(xy: VectorLike) -> Vector:
    """XYZ to UV."""

    u, v = xy_to_uv_1960(xy)
    return [u, v * (3 / 2)]


def uv_to_xy(uv: VectorLike) -> Vector:
    """XYZ to UV."""

    return uv_1960_to_xy([uv[0], uv[1] * (2 / 3)])


def xy_to_uv_1960(xy: VectorLike) -> Vector:
    """XYZ to UV."""

    x, y = xy
    denom = (12 * y - 2 * x + 3)
    if denom != 0:
        u = (4 * x) / denom
        v = (6 * y) / denom
    else:
        u = v = 0

    return [u, v]


def uv_1960_to_xy(uv: VectorLike) -> Vector:
    """XYZ to UV."""

    u, v = uv
    denom = (2 * u - 8 * v + 4)
    if denom != 0:
        x = (3 * u) / denom
        y = (2 * v) / denom
    else:
        x = y = 0

    return [x, y]


def pq_st2084_oetf(
    values: VectorLike,
    c1: float = C1,
    c2: float = C2,
    c3: float = C3,
    m1: float = M1,
    m2: float = M2
) -> Vector:
    """Perceptual quantizer (SMPTE ST 2084) - OETF."""

    adjusted = []
    for c in values:
        c = alg.npow(c / 10000, m1)
        r = (c1 + c2 * c) / (1 + c3 * c)
        adjusted.append(alg.npow(r, m2))
    return adjusted


def rgb_scale(vec: VectorLike) -> Vector:
    """
    Scale the RGB vector.

    If minimum is less than zero, behaves like min/max normalization.
    If minimum is not less than zero, behaves like maximum normalization.
    """

    # `(v - min_v)`
    w = min(vec)
    if w < 0.0:
        vec = [v - w for v in vec]

    # `(max_v - min_v)`
    m = max(vec)

    # `(v - min_v) / (max_v - min_v)`
    return [v / m if m else v for v in vec]


def pq_st2084_eotf(
    values: VectorLike,
    c1: float = C1,
    c2: float = C2,
    c3: float = C3,
    m1: float = M1,
    m2: float = M2
) -> Vector:
    """Perceptual quantizer (SMPTE ST 2084) - EOTF."""

    im1 = 1 / m1
    im2 = 1 / m2

    adjusted = []
    for c in values:
        c = alg.npow(c, im2)
        r = (c - c1) / (c2 - c3 * c)
        adjusted.append(10000 * alg.npow(r, im1))
    return adjusted


def xyz_to_absxyz(xyzd65: VectorLike, yw: float = YW) -> Vector:
    """XYZ to Absolute XYZ."""

    return [max(c * yw, 0) for c in xyzd65]


def absxyz_to_xyz(absxyzd65: VectorLike, yw: float = YW) -> Vector:
    """Absolute XYZ to XYZ."""

    return [max(c / yw, 0) for c in absxyzd65]


def constrain_hue(hue: float) -> float:
    """Constrain hue to [0, 360)."""

    return hue % 360 if not alg.is_nan(hue) else hue


def cmp_coords(c1: VectorLike, c2: VectorLike) -> bool:
    """Compare coordinates."""

    if len(c1) != len(c2):
        return False
    else:
        return all(map(lambda a, b: (math.isnan(a) and math.isnan(b)) or a == b, c1, c2))


def fmt_float(f: float, p: int = 0, percent: float = 0.0, offset: float = 0.0) -> str:
    """
    Set float precision and trim precision zeros.

    0: Round to whole integer
    -1: Full precision
    <positive number>: precision level
    """

    if alg.is_nan(f):
        return "none"

    value = alg.round_to((f + offset) / (percent * 0.01) if percent else f, p)
    if p == -1:
        p = 53
    return ('{{:{}{}g}}{}'.format('' if abs(value) >= 10 ** p else '.', p, '%' if percent else '')).format(value)



"""
Planckian curve.

https://en.wikipedia.org/wiki/Planckian_locus#The_Planckian_locus_in_the_XYZ_color_space
"""
from __future__ import annotations
import math
from ..types import VectorLike, Vector
from .. import util

# Constants for Planck's Law
# Precise calculation
# ```
# H = 6.62607015e-34  # Plank's constant: `m2 kg / s`
# C = 299792458  # Speed of light: `m / s`
# K = 1.380649e-23  # Boltzmann constant: `m2 kg s-2 K-1`
# C1 = 2 * math.pi * H * C ** 2  # First radiation constant
# C2 = (H * C) / K  # Second radiation constant
# ```
# ITS-90 Standard rounds to 6 decimal places
C1 = 3.741771e-16
C2 = 1.4388e-2


def temp_to_xy_planckian_locus(
    temp: float,
    cmfs: dict[int, tuple[float, float, float]],
    white: VectorLike,
    start: int = 360,
    end: int = 830,
    step: int = 5,
    c1: float = C1,
    c2: float = C2
) -> Vector:
    """
    Temperature to Planckian locus.

    https://en.wikipedia.org/wiki/Planckian_locus#The_Planckian_locus_in_the_XYZ_color_space
    """
    x = y = z = 0.0

    for wavelength in range(start, end + 1, step):
        m = c1 * (wavelength ** -5) * math.expm1((c2 * 1e9) / (wavelength * temp)) ** -1
        cmf = cmfs[wavelength]
        x += m * cmf[0]
        y += m * cmf[1]
        z += m * cmf[2]

    return util.xyz_to_xyY([x, y, z], white)[:-1]



"""Test IgPgTg."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
import pytest


class TestIgPgTg(util.ColorAssertsPyTest):
    """Test IgPgTg."""

    COLORS = [
        ('red', 'color(--igpgtg 0.54834 0.15366 0.43674)'),
        ('orange', 'color(--igpgtg 0.73238 0.0397 0.32108)'),
        ('yellow', 'color(--igpgtg 0.94415 -0.1422 0.39035)'),
        ('green', 'color(--igpgtg 0.42594 -0.18386 0.17055)'),
        ('blue', 'color(--igpgtg 0.3076 -0.26988 -0.40879)'),
        ('indigo', 'color(--igpgtg 0.22925 0.08712 -0.20068)'),
        ('violet', 'color(--igpgtg 0.68117 0.22553 -0.17411)'),
        ('white', 'color(--igpgtg 0.97415 0.00139 -0.00404)'),
        ('gray', 'color(--igpgtg 0.50619 0.00072 -0.0021)'),
        ('black', 'color(--igpgtg 0 0 0)'),
        # Test color
        ('color(--igpgtg 0.5 0.1 -0.1)', 'color(--igpgtg 0.5 0.1 -0.1)'),
        ('color(--igpgtg 0.5 0.1 -0.1 / 0.5)', 'color(--igpgtg 0.5 0.1 -0.1 / 0.5)'),
        ('color(--igpgtg 50% 50% -50% / 50%)', 'color(--igpgtg 0.5 0.5 -0.5 / 0.5)'),
        ('color(--igpgtg none none none / none)', 'color(--igpgtg none none none / none)'),
        # Test percent ranges
        ('color(--igpgtg 0% 0% 0%)', 'color(--igpgtg 0 0 0)'),
        ('color(--igpgtg 100% 100% 100%)', 'color(--igpgtg 1 1 1)'),
        ('color(--igpgtg -100% -100% -100%)', 'color(--igpgtg -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_igpgtg_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('igpgtg'), Color(color2))


class TestIgPgTgSerialize(util.ColorAssertsPyTest):
    """Test IgPgTg serialization."""

    COLORS = [
        # Test color
        ('color(--igpgtg 0.75 0.1 -0.1 / 0.5)', {}, 'color(--igpgtg 0.75 0.1 -0.1 / 0.5)'),
        # Test alpha
        ('color(--igpgtg 0.75 0.1 -0.1)', {'alpha': True}, 'color(--igpgtg 0.75 0.1 -0.1 / 1)'),
        ('color(--igpgtg 0.75 0.1 -0.1 / 0.5)', {'alpha': False}, 'color(--igpgtg 0.75 0.1 -0.1)'),
        # Test None
        ('color(--igpgtg none 0.1 -0.1)', {}, 'color(--igpgtg 0 0.1 -0.1)'),
        ('color(--igpgtg none 0.1 -0.1)', {'none': True}, 'color(--igpgtg none 0.1 -0.1)'),
        # Test Fit (not bound)
        ('color(--igpgtg 0.75 1.2 -0.1)', {}, 'color(--igpgtg 0.75 1.2 -0.1)'),
        ('color(--igpgtg 0.75 1.2 -0.1)', {'fit': False}, 'color(--igpgtg 0.75 1.2 -0.1)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestIgPgTgPoperties(util.ColorAsserts, unittest.TestCase):
    """Test IgPgTg."""

    def test_ig(self):
        """Test `ig`."""

        c = Color('color(--igpgtg 0.94415 -0.1422 0.39035)')
        self.assertEqual(c['ig'], 0.94415)
        c['ig'] = 0.2
        self.assertEqual(c['ig'], 0.2)

    def test_pg(self):
        """Test `pg`."""

        c = Color('color(--igpgtg 0.94415 -0.1422 0.39035)')
        self.assertEqual(c['pg'], -0.1422)
        c['pg'] = 0.1
        self.assertEqual(c['pg'], 0.1)

    def test_tg(self):
        """Test `tg`."""

        c = Color('color(--igpgtg 0.94415 -0.1422 0.39035)')
        self.assertEqual(c['tg'], 0.39035)
        c['tg'] = 0.1
        self.assertEqual(c['tg'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--igpgtg 0.94415 -0.1422 0.39035)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""Test CCT."""
import math
from coloraide.everything import ColorAll as Color
from coloraide import cmfs
from coloraide import cat
from . import util
import pytest
import unittest


class TestOhno2013Temp(util.ColorAssertsPyTest):
    """Test Ohno 2013."""

    @pytest.mark.parametrize('cct', list(range(1000, 100000, 100)))
    def test_cct(self, cct):
        """
        Test CCT methods.

        We test at a high resolution to make sure the spline implementation is working as good as without.
        """

        for duv in (-0.03, -0.02, -0.01, 0.0, 0.01, 0.02, 0.03):
            cct2, duv2 = Color.blackbody('xyz-d65', cct, duv, scale=False, method='ohno-2013').cct(method='ohno-2013')
            assert math.isclose(cct2, cct, rel_tol=(0.00001 if cct < 84000 else 0.0001), abs_tol=0.000001)
            assert math.isclose(duv2, duv, rel_tol=0.000001, abs_tol=0.000001)


class TestRobertson1968(util.ColorAssertsPyTest):
    """Test Robertson 1968."""

    @pytest.mark.parametrize('cct', list(range(1667, 100000, 1000)))
    def test_cct(self, cct):
        """Test CCT methods."""

        for duv in (-0.03, -0.02, -0.01, 0.0, 0.01, 0.02, 0.03):
            cct2, duv2 = Color.blackbody(
                'xyz-d65', cct, duv, scale=False, method='robertson-1968'
            ).cct(method='robertson-1968')
            assert math.isclose(cct2, cct, rel_tol=0.00001, abs_tol=0.00001)
            assert math.isclose(duv2, duv, rel_tol=0.00001, abs_tol=0.00001)


class TestOhno2013Color(util.ColorAssertsPyTest):
    """Test Ohno 2013."""

    @pytest.mark.parametrize(
        'color',
        [
            'green',
            'yellow',
            'orange',
            'white',
            'gray',
            'pink',

        ]
    )
    def test_cct(self, color):
        """Test CCT methods."""

        cct, duv = Color(color).cct()
        c = Color.blackbody('xyz-d65', cct, duv, scale=False)
        cct2, duv2 = c.cct()
        assert math.isclose(cct, cct2, rel_tol=0.00001, abs_tol=0.00001)
        assert math.isclose(duv, duv2, rel_tol=0.00001, abs_tol=0.00001)


class TestRobertson1968Color(util.ColorAssertsPyTest):
    """Test Robertson 1968."""

    @pytest.mark.parametrize(
        'color',
        [
            'green',
            'yellow',
            'orange',
            'white',
            'gray',
            'pink',

        ]
    )
    def test_cct(self, color):
        """Test CCT methods."""

        cct, duv = Color(color).cct(method='robertson-1968')
        c = Color.blackbody('xyz-d65', cct, duv, scale=False, method='robertson-1968')
        cct2, duv2 = c.cct(method='robertson-1968')
        assert math.isclose(cct, cct2, rel_tol=0.00001, abs_tol=0.00001)
        assert math.isclose(duv, duv2, rel_tol=0.00001, abs_tol=0.00001)


class TestCCTSpecificCases(util.ColorAsserts, unittest.TestCase):
    """Test CCT specific cases."""

    def test_bad_algorithm(self):
        """Test bad algorithm inputs."""

        with self.assertRaises(ValueError):
            Color.blackbody('srgb', 2000, method='bad')

        with self.assertRaises(ValueError):
            Color('blue').cct(method='bad')

    def test_outside_gamut(self):
        """Test temperature outside of gamut."""

        self.assertColorEqual(
            Color.blackbody('srgb-linear', 1500, scale_space='srgb-linear'),
            Color('color(srgb-linear 1 0.14966 0)')
        )

    def test_output_space(self):
        """Test output space."""

        self.assertColorEqual(
            Color.blackbody('display-p3', 2000, scale_space='srgb-linear'),
            Color('color(display-p3 0.93958 0.56696 0.2294)')
        )

    def test_normalization_space(self):
        """Test normalization space."""

        self.assertColorEqual(
            Color.blackbody('display-p3', 2000, scale_space='display-p3-linear'),
            Color('color(display-p3 1 0.60474 0.24668)')
        )

    def test_no_normalization_space(self):
        """Test normalization space."""

        self.assertColorEqual(
            Color.blackbody('xyz-d65', 2000, scale=False),
            Color('color(xyz-d65 1.2743 1 0.14517)')
        )

    def test_ohno_alternate_cmfs(self):
        """Test alternate CMFs."""

        from coloraide.temperature.ohno_2013 import Ohno2013

        class Custom(Color):
            CCT = 'ohno-2013'

        Custom.register(
            Ohno2013(cmfs.CIE_1964_10DEG, cat.WHITES['10deg']['D65']),
            overwrite=True
        )

        srgbl = Custom.blackbody('srgb-linear', 5000, scale_space='srgb-linear')
        self.assertColorEqual(srgbl, Color('color(srgb-linear 1 0.77908 0.61805)'))
        cct, duv = srgbl.cct()
        assert math.isclose(cct, 5000.005435878293, rel_tol=1e-11, abs_tol=1e-11)
        assert math.isclose(duv, 2.771032069214595e-08, rel_tol=1e-11, abs_tol=1e-11)

    def test_ohno_exact(self):
        """Test exact CMFs."""

        cct1, duv1 = Color('orange').cct(method='ohno-2013')
        assert math.isclose(cct1, 2424.1146637385255, rel_tol=1e-11, abs_tol=1e-11)
        assert math.isclose(duv1, 0.008069417642630583, rel_tol=1e-11, abs_tol=1e-11)
        cct2, duv2 = Color('orange').cct(method='ohno-2013', exact=True)
        assert math.isclose(cct2, 2424.1146637385255, rel_tol=1e-11, abs_tol=1e-11)
        assert math.isclose(duv2, 0.008069417642630583, rel_tol=1e-11, abs_tol=1e-11)

    def test_robertson_custom_table(self):
        """Test that we can customize Robertson 1968 table."""

        from coloraide.temperature import robertson_1968
        from coloraide import cmfs
        from coloraide import cat

        class Custom(Color):
            CCT = 'robertson-1968'

        Custom.register(
            robertson_1968.Robertson1968(cmfs.CIE_1964_10DEG, cat.WHITES['10deg']['D65']),
            overwrite=True
        )

        self.assertEqual(Custom.blackbody('xyz-d65', 1000, scale=False).cct(), [1000.0, 0.0])

    def test_robertson_sigfig(self):
        """Test that significant figure option."""

        from coloraide.temperature import robertson_1968
        from coloraide import cmfs
        from coloraide import cat

        class Custom(Color):
            CCT = 'robertson-1968'

        # Don't use significant figure rounding
        Custom.register(
            robertson_1968.Robertson1968(cmfs.CIE_1931_2DEG, cat.WHITES['2deg']['D65'], sigfig=0),
            overwrite=True
        )

        cct1, duv1 = Color('orange').cct()
        assert math.isclose(cct1, 2423.929632254136, rel_tol=1e-11, abs_tol=1e-11)
        assert math.isclose(duv1, 0.00811119780855307, rel_tol=1e-11, abs_tol=1e-11)
        cct2, duv2 = Custom('orange').cct()
        assert math.isclose(cct2, 2423.930481644873, rel_tol=1e-11, abs_tol=1e-11)
        assert math.isclose(duv2, 0.008112876273860207, rel_tol=1e-11, abs_tol=1e-11)



"""Provide filters as described by the https://www.w3.org/TR/filter-effects-1/."""
from __future__ import annotations
import math
from ..filters import Filter
from .. import algebra as alg
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


def linear_transfer(value: float, slope: float = 1.0, intercept: float = 0.0) -> float:
    """
    Linear transfer function.

    https://drafts.fxtf.org/filter-effects-1/#feFuncRElement
    """

    return value * slope + intercept


class Sepia(Filter):
    """Sepia filter."""

    NAME = 'sepia'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply a sepia filter to the color."""

        amount = 1 - alg.clamp(1 if amount is None else amount, 0, 1)

        m = [
            [0.393 + 0.607 * amount, 0.769 - 0.769 * amount, 0.189 - 0.189 * amount],
            [0.349 - 0.349 * amount, 0.686 + 0.314 * amount, 0.168 - 0.168 * amount],
            [0.272 - 0.272 * amount, 0.534 - 0.534 * amount, 0.131 + 0.869 * amount]
        ]

        color[:-1] = alg.dot(m, color[:-1], dims=alg.D2_D1)


class Grayscale(Filter):
    """Grayscale filter."""

    NAME = 'grayscale'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply a grayscale filter to the color."""

        amount = 1 - alg.clamp(1 if amount is None else amount, 0, 1)

        m = [
            [0.2126 + 0.7874 * amount, 0.7152 - 0.7152 * amount, 0.0722 - 0.0722 * amount],
            [0.2126 - 0.2126 * amount, 0.7152 + 0.2848 * amount, 0.0722 - 0.0722 * amount],
            [0.2126 - 0.2126 * amount, 0.7152 - 0.7152 * amount, 0.0722 + 0.9278 * amount]
        ]

        color[:-1] = alg.dot(m, color[:-1], dims=alg.D2_D1)


class Saturate(Filter):
    """Saturation filter."""

    NAME = 'saturate'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply a saturation filter to the color."""

        amount = alg.clamp(1 if amount is None else amount, 0)

        m = [
            [0.213 + 0.787 * amount, 0.715 - 0.715 * amount, 0.072 - 0.072 * amount],
            [0.213 - 0.213 * amount, 0.715 + 0.285 * amount, 0.072 - 0.072 * amount],
            [0.213 - 0.213 * amount, 0.715 - 0.715 * amount, 0.072 + 0.928 * amount]
        ]

        color[:-1] = alg.dot(m, color[:-1], dims=alg.D2_D1)


class Invert(Filter):
    """Grayscale filter."""

    NAME = 'invert'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply an invert filter."""

        amount = alg.clamp(1 if amount is None else amount, 0, 1)
        for e, c in enumerate(color[:-1]):
            color[e] = alg.lerp(amount, 1 - amount, c)


class Opacity(Filter):
    """Opacity filter."""

    NAME = 'opacity'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply an opacity filter."""

        amount = alg.clamp(1 if amount is None else amount, 0, 1)
        color[-1] = alg.lerp(0, amount, color[-1])


class Brightness(Filter):
    """Brightness filter."""

    NAME = 'brightness'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply a brightness filter."""

        amount = alg.clamp(1 if amount is None else amount, 0)
        for e, c in enumerate(color[:-1]):
            color[e] = linear_transfer(c, amount)


class Contrast(Filter):
    """Contrast filter."""

    NAME = 'contrast'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply a contrast filter."""

        amount = alg.clamp(1 if amount is None else amount, 0)
        for e, c in enumerate(color[:-1]):
            color[e] = linear_transfer(c, amount, (1 - amount) * 0.5)


class HueRotate(Filter):
    """Hue rotate filter."""

    NAME = 'hue-rotate'
    ALLOWED_SPACES = ('srgb-linear', 'srgb')

    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Apply a hue rotation filter."""

        rad = math.radians(0 if amount is None else amount)
        cos = math.cos(rad)
        sin = math.sin(rad)

        m = [
            [0.213 + cos * 0.787 - sin * 0.213, 0.715 - cos * 0.715 - sin * 0.715, 0.072 - cos * 0.072 + sin * 0.928],
            [0.213 - cos * 0.213 + sin * 0.143, 0.715 + cos * 0.285 + sin * 0.140, 0.072 - cos * 0.072 - sin * 0.283],
            [0.213 - cos * 0.213 - sin * 0.787, 0.715 - cos * 0.715 + sin * 0.715, 0.072 + cos * 0.928 + sin * 0.072]
        ]

        color[:-1] = alg.dot(m, color[:-1], dims=alg.D2_D1)



"""Convert the color."""
from __future__ import annotations
from .types import Vector
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .color import Color
    from .spaces import Space

# XYZ is the absolute base, meaning that XYZ is the final base in any conversion chain.
# This is a design expectation regardless of whether someone assigns a different base to XYZ or not.
ABSOLUTE_BASE = 'xyz-d65'


def calc_path_to_xyz(
    color: type[Color],
    space: str
) -> tuple[list[Space], dict[str, int]]:
    """
    Calculate the conversion path between a given color space and XYZ D65.

    We create two structures:

    1. A list containing the color space name in the conversion process from our target to XYZ D65.
    2. A mapping of color space names to the index in the color space name list.
    """

    obj = color.CS_MAP.get(space)
    if obj is None:
        raise ValueError("'{}' is not a valid color space".format(space))

    # Create a worse case conversion chain from XYZ to the target
    temp = obj
    count = 0
    from_color = []
    from_color_index = {}
    name = ''
    while name != ABSOLUTE_BASE:
        from_color.append(temp)
        name = temp.NAME
        from_color_index[name] = count
        temp = color.CS_MAP[temp.BASE]

        count += 1
        if count > color._MAX_CONVERT_ITERATIONS:  # pragma: no cover
            raise RuntimeError(
                'Conversion chain reached max size of {} and has terminated to avoid an infinite loop'.format(
                    count
                )
            )

    return from_color, from_color_index


def get_convert_chain(
    color: type[Color],
    space: Space,
    target: str
) -> list[tuple[Space, Space, int, bool]]:
    """
    Create a conversion chain.

    Each entry in the list will contain (from_space, to_space, direction, chromatic_adaptation_needed).
    Direction refers to whether conversions are moving to or from XYZ D65 as that will dictate whether
    `to_base` or `from_base` call method is used. If either the "from" or "to" color space is XYZ D65
    a chromatic adaptation will need to occur.
    """

    # Get the color space chain for the current space to XYZ
    from_color, from_color_index = calc_path_to_xyz(color, target)

    # Start building up the conversion chain.
    # The first stage builds up the chain towards XYZ D65.
    # If the color space we are converting to is not between
    # the current space and XYZ D65, nothing will get added.
    current = space
    chain = []  # type: list[tuple[Space, Space, int, bool]]
    if current.NAME != ABSOLUTE_BASE:
        count = 0
        while current.NAME not in from_color_index:

            # Get the "base space" (the space through which the current color converts to and from)
            base_space = color.CS_MAP[current.BASE]

            # Do we need to chromatically adapt towards XYZ D65?
            adapt = base_space.NAME == ABSOLUTE_BASE

            # Add conversion chain entry
            chain.append((current, base_space, 0, adapt))

            # The base space is now the current space
            current = base_space

            count += 1
            if count > color._MAX_CONVERT_ITERATIONS:  # pragma: no cover
                raise RuntimeError(
                    'Conversions reached max iteration of {} and has terminated to avoid an infinite loop'.format(
                        count
                    )
                )

    # If the chain still didn't resolve to the target space after the first stage,
    # build up the chain in the direction away from XYZ-D65.
    if current.NAME != target:
        # Start in the chain where the current color resides
        start = from_color_index[current.NAME] - 1

        # Do we need to chromatically adapt away from XYZ D65?
        adapt = current.NAME == ABSOLUTE_BASE

        # Moving away from XYZ D65, convert towards are desired target
        for index in range(start, -1, -1):
            base_space = current
            current = from_color[index]

            # Add the conversion chain entry
            chain.append((base_space, current, 1, adapt))

    return chain


def convert(color: Color, space: str) -> tuple[Space, Vector]:
    """Convert the color coordinates to the specified space."""

    # Grab the convert for the current space to the desired space
    # Result is cached for quicker future conversions.
    chain = color._get_convert_chain(color._space, space)  # type: ignore[attr-defined]

    # Get coordinates and convert NaN values to 0
    coords = color.coords(nans=False)

    # Navigate the conversion chain translating the coordinates along the way.
    # Perform chromatic adaption if needed (a conversion to or from XYZ D65).
    last = color._space
    for a, b, direction, adapt in chain:
        if direction and adapt:
            coords = color.chromatic_adaptation(
                a.WHITE,
                b.WHITE,
                coords
            )

        coords = b.from_base(coords) if direction else a.to_base(coords)
        if not direction and adapt:
            coords = color.chromatic_adaptation(
                a.WHITE,
                b.WHITE,
                coords
            )
        last = b

    return last, coords



"""
Luv class.

https://en.wikipedia.org/wiki/CIELuv
"""
from __future__ import annotations
from ..spaces import Space, Labish
from ..cat import WHITES
from ..channels import Channel, FLG_MIRROR_PERCENT
from .lab import KAPPA, EPSILON, KE, ACHROMATIC_THRESHOLD
from .. import util
from .. import algebra as alg
from ..types import Vector


def xyz_to_luv(xyz: Vector, white: tuple[float, float]) -> Vector:
    """XYZ to Luv."""

    u, v = util.xy_to_uv(util.xyz_to_xyY(xyz, white)[:2])
    w_xyz = util.xy_to_xyz(white)
    ur, vr = util.xy_to_uv(white)

    yr = xyz[1] / w_xyz[1]
    l = 116 * alg.nth_root(yr, 3) - 16 if yr > EPSILON else KAPPA * yr

    return [
        l,
        13 * l * (u - ur),
        13 * l * (v - vr),
    ]


def luv_to_xyz(luv: Vector, white: tuple[float, float]) -> Vector:
    """Luv to XYZ."""

    l, u, v = luv
    w_xyz = util.xy_to_xyz(white)
    ur, vr = util.xy_to_uv(white)

    if l != 0:
        up = (u / (13 * l)) + ur
        vp = (v / (13 * l)) + vr
    else:
        up = vp = 0

    y = w_xyz[1] * (((l + 16) / 116) ** 3 if l > KE else l / KAPPA)

    if vp != 0:
        x = y * ((9 * up) / (4 * vp))
        z = y * ((12 - 3 * up - 20 * vp) / (4 * vp))
    else:
        x = z = 0

    return [x, y, z]


class Luv(Labish, Space):
    """Luv class."""

    BASE = "xyz-d65"
    NAME = "luv"
    SERIALIZE = ("--luv",)
    CHANNELS = (
        Channel("l", 0.0, 100.0),
        Channel("u", -215.0, 215.0, flags=FLG_MIRROR_PERCENT),
        Channel("v", -215.0, 215.0, flags=FLG_MIRROR_PERCENT)
    )
    CHANNEL_ALIASES = {
        "lightness": "l"
    }
    WHITE = WHITES['2deg']['D65']

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return coords[0] == 0.0 or alg.rect_to_polar(coords[1], coords[2])[0] < ACHROMATIC_THRESHOLD

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ D50 from Luv."""

        return luv_to_xyz(coords, self.WHITE)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ D50 to Luv."""

        return xyz_to_luv(coords, self.WHITE)



"""Everything and the kitchen sink."""
from __future__ import annotations
from .color import Color as Base
from .spaces.rec2100_pq import Rec2100PQ
from .spaces.rec2100_hlg import Rec2100HLG
from .spaces.jzazbz import Jzazbz
from .spaces.jzczhz import JzCzhz
from .spaces.ictcp import ICtCp
from .spaces.din99o import DIN99o
from .spaces.lch99o import LCh99o
from .spaces.luv import Luv
from .spaces.lchuv import LChuv
from .spaces.hsluv import HSLuv
from .spaces.hpluv import HPLuv
from .spaces.okhsl import Okhsl
from .spaces.okhsv import Okhsv
from .spaces.hsi import HSI
from .spaces.ipt import IPT
from .spaces.igpgtg import IgPgTg
from .spaces.cmy import CMY
from .spaces.cmyk import CMYK
from .spaces.xyy import xyY
from .spaces.xyb import XYB
from .spaces.hunter_lab import HunterLab
from .spaces.prismatic import Prismatic
from .spaces.rlab import RLAB
from .spaces.orgb import oRGB
from .spaces.aces2065_1 import ACES20651
from .spaces.acescg import ACEScg
from .spaces.acescc import ACEScc
from .spaces.acescct import ACEScct
from .spaces.cam16 import CAM16
from .spaces.cam16_jmh import CAM16JMh
from .spaces.cam16_ucs import CAM16UCS, CAM16LCD, CAM16SCD
from .spaces.hct import HCT
from .spaces.ucs import UCS
from .spaces.rec709 import Rec709
from .distance.delta_e_itp import DEITP
from .distance.delta_e_99o import DE99o
from .distance.delta_e_z import DEZ
from .distance.delta_e_cam16 import DECAM16
from .distance.delta_e_hct import DEHCT
from .gamut.fit_hct_chroma import HCTChroma
from .interpolate.catmull_rom import CatmullRom
from .contrast.lstar import LstarContrast
from .cat import VonKries, XYZScaling, CAT02, CMCCAT97, Sharp, CMCCAT2000, CAT16
from .color import ColorMatch
from .interpolate import stop, hint
from .algebra import NaN

__all__ = ('ColorAll', 'ColorMatch', 'stop', 'hint', 'NaN')


class ColorAll(Base):
    """Color with all plugins."""


ColorAll.register(
    [
        # Spaces
        Rec709(),
        Rec2100PQ(),
        Rec2100HLG(),
        Jzazbz(),
        JzCzhz(),
        ICtCp(),
        DIN99o(),
        LCh99o(),
        Luv(),
        LChuv(),
        Okhsl(),
        Okhsv(),
        HSLuv(),
        HPLuv(),
        HSI(),
        IPT(),
        IgPgTg(),
        CMY(),
        CMYK(),
        xyY(),
        XYB(),
        HunterLab(),
        Prismatic(),
        RLAB(),
        oRGB(),
        ACES20651(),
        ACEScg(),
        ACEScc(),
        ACEScct(),
        CAM16(),
        CAM16JMh(),
        CAM16UCS(),
        CAM16SCD(),
        CAM16LCD(),
        HCT(),
        UCS(),

        # Delta E
        DEITP(),
        DE99o(),
        DEZ(),
        DECAM16(),
        DEHCT(),

        # Gamut Mapping
        HCTChroma(),

        # CAT
        VonKries(),
        XYZScaling(),
        CAT02(),
        CMCCAT97(),
        Sharp(),
        CMCCAT2000(),
        CAT16(),

        # Interpolation
        CatmullRom(),

        # Contrast
        LstarContrast()
    ]
)



"""Test XYZ D65 library."""
import unittest
from . import util
from coloraide import Color
import pytest


class TestXYZD65(util.ColorAssertsPyTest):
    """Test XYZ D65."""

    COLORS = [
        ('red', 'color(xyz-d65 0.41239 0.21264 0.01933)'),
        ('orange', 'color(xyz-d65 0.54694 0.48173 0.06418)'),
        ('yellow', 'color(xyz-d65 0.76998 0.92781 0.13853)'),
        ('green', 'color(xyz-d65 0.07719 0.15438 0.02573)'),
        ('blue', 'color(xyz-d65 0.18048 0.07219 0.95053)'),
        ('indigo', 'color(xyz-d65 0.0693 0.03108 0.21355)'),
        ('violet', 'color(xyz-d65 0.58672 0.40317 0.85583)'),
        ('white', 'color(xyz-d65 0.95046 1 1.0891)'),
        ('gray', 'color(xyz-d65 0.20517 0.21586 0.23508)'),
        ('black', 'color(xyz-d65 0 0 0)'),
        # Test alternate CSS identifier
        ('color(xyz 0.95046 1 1.0891)', 'color(xyz-d65 0.95046 1 1.0891)'),
        # Test CSS color
        ('color(xyz-d65 0 0.50196 0)', 'color(xyz-d65 0 0.50196 0)'),
        ('color(xyz-d65 0 0.50196 0 / 0.5)', 'color(xyz-d65 0 0.50196 0 / 0.5)'),
        ('color(xyz-d65 50% 50% 50% / 50%)', 'color(xyz-d65 0.5 0.5 0.5 / 0.5)'),
        ('color(xyz-d65 none none none / none)', 'color(xyz-d65 none none none / none)'),
        # Test range
        ('color(xyz-d65 0% 0% 0%)', 'color(xyz-d65 0 0 0)'),
        ('color(xyz-d65 100% 100% 100%)', 'color(xyz-d65 1 1 1)'),
        ('color(xyz-d65 -100% -100% -100%)', 'color(xyz-d65 -1 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('xyz-d65'), Color(color2))


class TestXYZD65Serialize(util.ColorAssertsPyTest):
    """Test XYZ D65 serialization."""

    COLORS = [
        # Test color
        ('color(xyz-d65 0 0.3 0.75 / 0.5)', {}, 'color(xyz-d65 0 0.3 0.75 / 0.5)'),
        # Test alpha
        ('color(xyz-d65 0 0.3 0.75)', {'alpha': True}, 'color(xyz-d65 0 0.3 0.75 / 1)'),
        ('color(xyz-d65 0 0.3 0.75 / 0.5)', {'alpha': False}, 'color(xyz-d65 0 0.3 0.75)'),
        # Test None
        ('color(xyz-d65 none 0.3 0.75)', {}, 'color(xyz-d65 0 0.3 0.75)'),
        ('color(xyz-d65 none 0.3 0.75)', {'none': True}, 'color(xyz-d65 none 0.3 0.75)'),
        # Test Fit (not bound)
        ('color(xyz-d65 1.2 0.2 0)', {}, 'color(xyz-d65 1.2 0.2 0)'),
        ('color(xyz-d65 1.2 0.2 0)', {'fit': False}, 'color(xyz-d65 1.2 0.2 0)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestXYZD65Properties(util.ColorAsserts, unittest.TestCase):
    """Test XYZ D65."""

    def test_names(self):
        """Test RGB-ish names."""

        self.assertEqual(Color('color(xyz 0.1 0.2 0.3 / 1)')._space.names(), ('x', 'y', 'z'))

    def test_indexes(self):
        """Test RGB-ish names."""

        self.assertEqual(Color('color(xyz 0.1 0.2 0.3 / 1)')._space.indexes(), [0, 1, 2])

    def test_x(self):
        """Test `x`."""

        c = Color('color(xyz 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['x'], 0.1)
        c['x'] = 0.2
        self.assertEqual(c['x'], 0.2)

    def test_y(self):
        """Test `y`."""

        c = Color('color(xyz 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['y'], 0.2)
        c['y'] = 0.1
        self.assertEqual(c['y'], 0.1)

    def test_z(self):
        """Test `z`."""

        c = Color('color(xyz 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['z'], 0.3)
        c['z'] = 0.1
        self.assertEqual(c['z'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(xyz 0.1 0.2 0.3 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)



"""DIN99o LCh class."""
from __future__ import annotations
from ..cat import WHITES
from .lch import LCh
from ..channels import Channel, FLG_ANGLE


class LCh99o(LCh):
    """DIN99o LCh class."""

    BASE = 'din99o'
    NAME = "lch99o"
    SERIALIZE = ("--lch99o",)
    WHITE = WHITES['2deg']['D65']
    CHANNELS = (
        Channel("l", 0.0, 100.0),
        Channel("c", 0.0, 60.0, limit=(0.0, None)),
        Channel("h", 0.0, 360.0, flags=FLG_ANGLE)
    )



"""Test JzCzhz library."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color
from coloraide import NaN
import pytest


class TestJzCzhz(util.ColorAssertsPyTest):
    """Test JzCzhz."""

    COLORS = [
        ('red', 'color(--jzczhz 0.13438 0.16252 43.502)'),
        ('orange', 'color(--jzczhz 0.16937 0.12698 75.776)'),
        ('yellow', 'color(--jzczhz 0.2096 0.1378 102)'),
        ('green', 'color(--jzczhz 0.09203 0.10932 132.99)'),
        ('blue', 'color(--jzczhz 0.09577 0.19029 257.61)'),
        ('indigo', 'color(--jzczhz 0.06146 0.10408 287.05)'),
        ('violet', 'color(--jzczhz 0.16771 0.08468 319.37)'),
        ('white', 'color(--jzczhz 0.22207 0.0002 none)'),
        ('gray', 'color(--jzczhz 0.11827 0.00014 none)'),
        ('black', 'color(--jzczhz 0 0 none)'),
        # Test color
        ('color(--jzczhz 1 0.3 270)', 'color(--jzczhz 1 0.3 270)'),
        ('color(--jzczhz 1 0.3 270 / 0.5)', 'color(--jzczhz 1 0.3 270 / 0.5)'),
        ('color(--jzczhz 50% 50% 50% / 50%)', 'color(--jzczhz 0.5 0.25 180 / 0.5)'),
        ('color(--jzczhz none none none / none)', 'color(--jzczhz none none none / none)'),
        # Test percent ranges
        ('color(--jzczhz 0% 0% 0%)', 'color(--jzczhz 0 0 0)'),
        ('color(--jzczhz 100% 100% 100%)', 'color(--jzczhz 1 0.5 360 / 1)'),
        ('color(--jzczhz -100% -100% -100%)', 'color(--jzczhz 0 0 -360 / 1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('jzczhz'), Color(color2))


class TestJzCzhzSerialize(util.ColorAssertsPyTest):
    """Test JzCzhz serialization."""

    COLORS = [
        # Test color
        ('color(--jzczhz 0.75 0.3 50 / 0.5)', {}, 'color(--jzczhz 0.75 0.3 50 / 0.5)'),
        # Test alpha
        ('color(--jzczhz 0.75 0.3 50)', {'alpha': True}, 'color(--jzczhz 0.75 0.3 50 / 1)'),
        ('color(--jzczhz 0.75 0.3 50 / 0.5)', {'alpha': False}, 'color(--jzczhz 0.75 0.3 50)'),
        # Test None
        ('color(--jzczhz none 0.3 50)', {}, 'color(--jzczhz 0 0.3 50)'),
        ('color(--jzczhz none 0.3 50)', {'none': True}, 'color(--jzczhz none 0.3 50)'),
        # Test Fit (not bound)
        ('color(--jzczhz 0.75 0.6 50)', {}, 'color(--jzczhz 0.75 0.6 50)'),
        ('color(--jzczhz 0.75 0.6 50)', {'fit': False}, 'color(--jzczhz 0.75 0.6 50)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestJzCzhzProperties(util.ColorAsserts, unittest.TestCase):
    """Test JzCzhz."""

    def test_jz(self):
        """Test `jz`."""

        c = Color('color(--jzczhz 0.22 0.5 270 / 1)')
        self.assertEqual(c['jz'], 0.22)
        c['jz'] = 0.2
        self.assertEqual(c['jz'], 0.2)

    def test_cz(self):
        """Test `chroma`."""

        c = Color('color(--jzczhz 0.22 0.5 270 / 1)')
        self.assertEqual(c['chroma'], 0.5)
        c['chroma'] = 0.1
        self.assertEqual(c['chroma'], 0.1)

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--jzczhz 0.22 0.5 270 / 1)')
        self.assertEqual(c['hue'], 270)
        c['hue'] = 0.1
        self.assertEqual(c['hue'], 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--jzczhz 0.22 0.5 270 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)

    def test_hue_name(self):
        """Test `hue_name`."""

        c = Color('color(--jzczhz 0.22 0.5 270 / 1)')
        self.assertEqual(c._space.hue_name(), 'hz')


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('jzczhz', [90, 50, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('color(--jzczhz 90% 0 none / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_chroma(self):
        """Test minimum saturation."""

        c = Color(Color('white').convert('jzczhz').to_string(precision=6)).normalize()
        self.assertTrue(c.is_nan('hue'))

        c = Color(Color('gray').convert('jzczhz').to_string(precision=6)).normalize()
        self.assertTrue(c.is_nan('hue'))

        c = Color(Color('darkgray').convert('jzczhz').to_string(precision=6)).normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to OkLCh with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('jzczhz')
                self.assertTrue(color2.is_nan('hue'))


class TestQuirks(util.ColorAsserts, unittest.TestCase):
    """Test quirks."""

    def test_chorma_less_than_zero_convert(self):
        """Test handling of negative chroma when converting to Jzazbz."""

        self.assertColorEqual(
            Color('color(--jzczhz 90% -10 120 / 1)').convert('jzazbz'), Color('color(--jzazbz 0.9 0 0)')
        )


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('#222222').convert('jzczhz').is_achromatic(), True)
        self.assertEqual(Color('#222222').convert('jzczhz').set('cz', lambda x: x + 1e-8).is_achromatic(), True)
        self.assertEqual(Color('srgb', [5.2, 5.2, 5.2]).convert('jzczhz').is_achromatic(), True)
        self.assertEqual(Color('jzczhz', [NaN, 0.00001, 270]).is_achromatic(), True)
        self.assertEqual(Color('jzczhz', [0, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('jzczhz', [0, 0.5, 270]).is_achromatic(), True)
        self.assertEqual(Color('pink').convert('jzczhz').is_achromatic(), False)
        self.assertEqual(Color('jzczhz', [NaN, 0.5, 270]).is_achromatic(), True)
        self.assertEqual(Color('jzczhz', [0.2, NaN, 270]).is_achromatic(), False)
        self.assertEqual(Color('jzczhz', [NaN, NaN, 270]).is_achromatic(), True)



"""
Calculate ICtCp matrices.

https://professional.dolby.com/siteassets/pdfs/ictcp_dolbywhitepaper_v071.pdf
"""
import sys
import os

sys.path.insert(0, os.getcwd())

from coloraide import algebra as alg  # noqa: E402


a = [
    [0.92, 0.04, 0.04],
    [0.04, 0.92, 0.04],
    [0.04, 0.04, 0.92]
]

b = [
    [0.4002, 0.7076, -0.0808],
    [-0.2263, 1.1653, 0.0457],
    [0, 0, 0.9182]
]

# XYZ to LMS
m1 = alg.dot(a, b)

# LMS to ICtCp
m2 = [
    [2048, 2048, 0],
    [6610, -13613, 7003],
    [17933, -17390, -543]
]

m2 = alg.divide(m2, 4096)


def pprint(value):
    """Print the matrix."""
    print('[', end='')
    first = True
    for v in value:
        if first:
            first = False
        else:
            print(',\n ', end='')
        print(v, end='')
    print(']')


if __name__ == "__main__":
    print('===== XYZ to LMS =====')
    pprint(m1)
    print('===== LMS to XYZ =====')
    pprint(alg.inv(m1))
    print('===== PQ LMS to ICtCp =====')
    pprint(m2)
    print('===== ICtCp to PQ LMS =====')
    pprint(alg.inv(m2))



"""Snapshot tests for HSLuv and HPLuv."""
import unittest
from .. import util
from coloraide.everything import ColorAll as Color
import json
import os


class TestHSLuvSnapshot(util.ColorAsserts, unittest.TestCase):
    """Run tests against the official snapshot."""

    def setUp(self):
        """Load snapshot."""

        snapshot = os.path.join(os.path.dirname(__file__), 'snapshot-rev4.json')
        with open(snapshot, 'r') as f:
            self.snapshot = json.loads(f.read())

    def test_snapshot(self):
        """
        Test the snapshot.

        https://github.com/hsluv/hsluv/tree/master/snapshots.

        The snapshot contains a number of unnecessary tests, at least for our purposes. It is sufficient to verify that
        our path from sRGB to HSLuv and HPLuv is correct, both forward and backwards, as our conversion path is noted
        below:

            ```
            srgb -> srgb-linear -> xyz-d65 -> luv -> lchuv -> hsluv
            srgb -> srgb-linear -> xyz-d65 -> luv -> lchuv -> hpluv
            ```

        Verifying each step in the chain makes no sense. If we are able to get through this entire chain and the sRGB
        to HSLuv checks out, both forward and backwards through the path, then we can consider our implementation
        successful. Any failure in the chain would cause errors to propagate up that would cause the test to fail. It
        would only be useful to test each point in the chain if we were trying to identify where in the chain breakage
        occurred.
        """

        for rgb, colors in self.snapshot.items():
            self.assertColorEqual(
                Color(rgb).convert('hsluv'),
                Color('hsluv', colors['hsluv'])
            )

            self.assertColorEqual(
                Color('hsluv', colors['hsluv']).convert('srgb'),
                Color('srgb', colors['rgb'])
            )

            self.assertColorEqual(
                Color(rgb).convert('hpluv'),
                Color('hpluv', colors['hpluv'])
            )

            self.assertColorEqual(
                Color('hpluv', colors['hpluv']).convert('srgb'),
                Color('srgb', colors['rgb'])
            )



"""Harmony diagram."""
import sys
import os
import matplotlib.pyplot as plt
import argparse
import math

sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color
from tools.slice_diagram import plot_slice  # noqa: E402


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='gamut_diagrams', description='Demonstrate gamut mapping.')
    parser.add_argument('--space', '-s', default='hsl', help="Space to interpolate in.")
    parser.add_argument('--color', '-c', help="Seed color.")
    parser.add_argument('--harmony', '-n', default='complement', help="Color harmony to use.")
    parser.add_argument('--gamut', '-g', default="srgb", help='Gamut to evaluate the color in (default is sRGB).')
    parser.add_argument('--map-colors', '-m', action='store_true', help="Gamut map colors to be within the gamut.")
    parser.add_argument('--title', '-t', default='', help="Provide a title for the diagram.")
    parser.add_argument('--yaxis', '-y', help="The channel to plot on Y axis (other than hue or l) 'name:min:max'.")
    parser.add_argument('--resolution', '-r', default="800", help="How densely to render the figure.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')

    args = parser.parse_args()

    c1 = Color(args.color).convert(args.space)

    names = set(c1._space.CHANNELS + tuple(c1._space.CHANNEL_ALIASES.keys()))

    for c_name in ('lightness', 'intensity', 'l', 'i'):
        if c_name in names:
            c_value = c1[c_name]
            break
    else:
        raise ValueError('Cannot find a lightness equivalence')

    if 'hue' not in names:
        raise ValueError('Expected a color space with a "hue" attribute.')

    if not args.title:
        title = "Color Harmony '{}'' in Color Space '{}' in Gamut '{}'".format(args.harmony, args.space, args.gamut)
    else:
        title = args.title

    subtitle = 'Initial Color: {}'.format(c1.to_string())

    plot_slice(
        args.space,
        '{}:{}'.format(c_name, c_value),
        'hue:0:360',
        args.yaxis,
        gamut=args.gamut,
        resolution=int(args.resolution),
        title=title,
        subtitle=subtitle,
        dark=args.dark,
        polar=True
    )

    # Get the actual indexes of the specified channels
    name2 = args.yaxis.split(":", 1)[0]
    name2 = c1._space.CHANNEL_ALIASES.get(name2, name2)

    colors = c1.harmony(args.harmony, space=args.space)
    if args.map_colors:
        [c.fit(args.gamut) for c in colors]

    for c in colors:
        plt.scatter(
            math.radians(c.get('hue')),
            c.get(name2),
            marker="o",
            color=c.convert('srgb').to_string(hex=True),
            edgecolor='black',
            s=8 ** 2,
            zorder=100
        )
        xs = [0, math.radians(c.get('hue'))]
        ys = [0, c.get(name2)]

        plt.plot(
            xs,
            ys,
            color='black',
            marker="",
            linewidth=1.5,
            markersize=2,
            antialiased=True
        )

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""Test CAM16 JMh."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
from coloraide.spaces.cam16_jmh import cam16_to_xyz_d65, CAM16JMh
from collections import namedtuple
import pytest

CAM16Coords = namedtuple("CAM16Coords", "J C h s Q M H")


class TestCAM16JMh(util.ColorAssertsPyTest):
    """Test CAM16 JMh."""

    COLORS = [
        ('red', 'color(--cam16-jmh 46.026 81.254 27.393)'),
        ('orange', 'color(--cam16-jmh 68.056 43.51 71.293)'),
        ('yellow', 'color(--cam16-jmh 94.682 54.5 111.15)'),
        ('green', 'color(--cam16-jmh 33.976 50.934 142.3)'),
        ('blue', 'color(--cam16-jmh 25.066 62.442 282.75)'),
        ('indigo', 'color(--cam16-jmh 16.046 43.278 310.9)'),
        ('violet', 'color(--cam16-jmh 63.507 46.779 331.39)'),
        ('white', 'color(--cam16-jmh 100 2.2369 209.53)'),
        ('gray', 'color(--cam16-jmh 43.042 1.467 209.54)'),
        ('black', 'color(--cam16-jmh 0 0 209.54)'),
        # Test color
        ('color(--cam16-jmh 50 30 270)', 'color(--cam16-jmh 50 30 270)'),
        ('color(--cam16-jmh 50 30 270 / 0.5)', 'color(--cam16-jmh 50 30 270 / 0.5)'),
        ('color(--cam16-jmh 50% 50% 50% / 50%)', 'color(--cam16-jmh 50 52.5 180 / 0.5)'),
        ('color(--cam16-jmh none none none / none)', 'color(--cam16-jmh none none none / none)'),
        # Test percent ranges
        ('color(--cam16-jmh 0% 0% 0%)', 'color(--cam16-jmh 0 0 0)'),
        ('color(--cam16-jmh 100% 100% 100%)', 'color(--cam16-jmh 100 105 360 / 1)'),
        ('color(--cam16-jmh -100% -100% -100%)', 'color(--cam16-jmh 0 0 -360 / 1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('cam16-jmh'), Color(color2))


class TestCAM16JMhPoperties(util.ColorAsserts, unittest.TestCase):
    """Test CAM16 JMh properties."""

    def test_names(self):
        """Test LCh-ish names."""

        self.assertEqual(Color('color(--cam16-jmh 97.139 75.504 111.05)')._space.names(), ('j', 'm', 'h'))

    def test_h(self):
        """Test `h`."""

        c = Color('color(--cam16-jmh 97.139 75.504 111.05)')
        self.assertEqual(c['h'], 111.05)
        c['h'] = 270
        self.assertEqual(c['h'], 270)

    def test_c(self):
        """Test `m`."""

        c = Color('color(--cam16-jmh 97.139 75.504 111.05)')
        self.assertEqual(c['m'], 75.504)
        c['m'] = 30
        self.assertEqual(c['m'], 30)

    def test_j(self):
        """Test `j`."""

        c = Color('color(--cam16-jmh 97.139 75.504 111.05)')
        self.assertEqual(c['j'], 97.139)
        c['j'] = 50
        self.assertEqual(c['j'], 50)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--cam16-jmh 97.139 75.504 111.05)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('cam16-jmh', [30, 20, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('color(--cam16-jmh 30 20 none / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_chroma(self):
        """Test minimum saturation."""

        c = Color(Color('white').convert('cam16-jmh').to_string()).normalize()
        self.assertTrue(c.is_nan('hue'))

        c = Color(Color('gray').convert('cam16-jmh').to_string()).normalize()
        self.assertTrue(c.is_nan('hue'))

        c = Color(Color('darkgray').convert('cam16-jmh').to_string()).normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to CAM16 JMh with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('cam16-jmh')
                self.assertTrue(color2.is_nan('hue'))


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('#222222').convert('cam16-jmh').is_achromatic(), True)
        self.assertEqual(
            Color('srgb', [0.000000001] * 3).convert('cam16-jmh').set('m', lambda x: x + 1e-8).is_achromatic(),
            True
        )
        self.assertEqual(Color('cam16-jmh', [NaN, 0.00001, 270]).is_achromatic(), True)
        self.assertEqual(Color('cam16-jmh', [0, NaN, 270]).is_achromatic(), True)
        self.assertEqual(Color('cam16-jmh', [0, 50, 270]).is_achromatic(), True)
        self.assertEqual(Color('cam16-jmh', [NaN, 50, 270]).is_achromatic(), True)
        self.assertEqual(Color('cam16-jmh', [20, NaN, 270]).is_achromatic(), False)
        self.assertEqual(Color('cam16-jmh', [NaN, NaN, 270]).is_achromatic(), True)


class TestCAM16ApperanceModel(util.ColorAsserts, unittest.TestCase):
    """Test CAM16 appearance model."""

    COORDS = CAM16Coords(
        45.33435136131785, 45.26195932727762, 258.92464993097565,
        62.67686398624793, 83.29355481993107, 32.720950777696196, 310.5279473979526
    )

    def test_no_lightness(self):
        """Test conversion failure when no equivalent lightness."""

        with self.assertRaises(ValueError):
            cam16_to_xyz_d65(C=self.COORDS.C, h=self.COORDS.h, env=CAM16JMh.ENV)

    def test_no_chroma(self):
        """Test conversion failure when no equivalent chroma."""

        with self.assertRaises(ValueError):
            cam16_to_xyz_d65(J=self.COORDS.J, h=self.COORDS.h, env=CAM16JMh.ENV)

    def test_no_hue(self):
        """Test conversion failure when no equivalent hue."""

        with self.assertRaises(ValueError):
            cam16_to_xyz_d65(J=self.COORDS.J, C=self.COORDS.C, env=CAM16JMh.ENV)

    def test_no_environment(self):
        """Test no test no environment."""

        with self.assertRaises(ValueError):
            cam16_to_xyz_d65(J=self.COORDS.J, C=self.COORDS.C, h=self.COORDS.h)

    def test_lightness_convert(self):
        """Test convert lightness."""

        for a, b in zip(
            cam16_to_xyz_d65(J=self.COORDS.J, C=self.COORDS.C, h=self.COORDS.h, env=CAM16JMh.ENV),
            cam16_to_xyz_d65(Q=self.COORDS.Q, C=self.COORDS.C, h=self.COORDS.h, env=CAM16JMh.ENV)
        ):
            self.assertCompare(a, b, 14)

    def test_chroma_convert(self):
        """Test convert chroma."""

        for a, b in zip(
            cam16_to_xyz_d65(J=self.COORDS.J, C=self.COORDS.C, h=self.COORDS.h, env=CAM16JMh.ENV),
            cam16_to_xyz_d65(Q=self.COORDS.Q, s=self.COORDS.s, h=self.COORDS.h, env=CAM16JMh.ENV)
        ):
            self.assertCompare(a, b, 14)

        for a, b in zip(
            cam16_to_xyz_d65(J=self.COORDS.J, C=self.COORDS.C, h=self.COORDS.h, env=CAM16JMh.ENV),
            cam16_to_xyz_d65(Q=self.COORDS.Q, M=self.COORDS.M, h=self.COORDS.h, env=CAM16JMh.ENV)
        ):
            self.assertCompare(a, b, 14)

    def test_hue_convert(self):
        """Test convert hue."""

        for a, b in zip(
            cam16_to_xyz_d65(J=self.COORDS.J, C=self.COORDS.C, h=self.COORDS.h, env=CAM16JMh.ENV),
            cam16_to_xyz_d65(J=self.COORDS.J, C=self.COORDS.C, H=self.COORDS.H, env=CAM16JMh.ENV)
        ):
            self.assertCompare(a, b, 14)



"""Test HSL library."""
from coloraide import NaN
import unittest
from . import util
from coloraide import Color
import pytest


class TestHSL(util.ColorAssertsPyTest):
    """Test HSL."""

    COLORS = [
        ('red', 'color(--hsl 0 1 0.5)'),
        ('orange', 'color(--hsl 38.824 1 0.5)'),
        ('yellow', 'color(--hsl 60 1 0.5)'),
        ('green', 'color(--hsl 120 1 0.25098)'),
        ('blue', 'color(--hsl 240 1 0.5)'),
        ('indigo', 'color(--hsl 274.62 1 0.2549)'),
        ('violet', 'color(--hsl 300 0.76056 0.72157)'),
        ('white', 'color(--hsl none 0 1)'),
        ('gray', 'color(--hsl none 0 0.50196)'),
        ('black', 'color(--hsl none 0 0)'),
        # Test legacy CSS
        ('hsl(270, 30%, 50%)', 'color(--hsl 270 0.3 0.5)'),
        ('hsl(270, 30%, 50%, 0.5)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsl(270, 30%, 50%, 50%)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsla(270, 30%, 50%)', 'color(--hsl 270 0.3 0.5)'),
        ('hsla(270, 30%, 50%, 0.5)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsla(270, 30%, 50%, 50%)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsl(none, 30%, 50%)', None),
        ('hsl(50%, 50%, 50%)', None),
        ('hsla(none, 30%, 50%)', None),
        ('hsla(50%, 50%, 50%)', None),
        ('hsl(50%, 50%)', None),
        ('hsl(270, 30, 50)', None),
        ('hsl(270, 30%, 50%, 3deg)', None),
        # Test CSS
        ('hsl(270 30% 50%)', 'color(--hsl 270 0.3 0.5)'),
        ('hsl(270 30% 50% / 0.5)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsl(270 30% 50% / 50%)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsl(none none none / none)', 'color(--hsl none none none / none)'),
        ('hsla(270 30% 50%)', 'color(--hsl 270 0.3 0.5)'),
        ('hsla(270 30% 50% / 0.5)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsla(270 30% 50% / 50%)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsla(none none none / none)', 'color(--hsl none none none / none)'),
        ('hsl(270 30 50)', 'color(--hsl 270 0.3 0.5)'),
        ('hsl(270 30% 50)', 'color(--hsl 270 0.3 0.5)'),
        ('hsla(270 30 50 / 0.5)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('hsl(50% 30 50)', None),
        ('hsl(270 30deg 50)', None),
        ('hsl(270 30% 50% 30%)', None),
        ('hsl(270 30% 50% / 3deg)', None),
        # Test CSS color
        ('color(--hsl 270 0.3 0.5)', 'color(--hsl 270 0.3 0.5)'),
        ('color(--hsl 270 0.3 0.5 / 0.5)', 'color(--hsl 270 0.3 0.5 / 0.5)'),
        ('color(--hsl 50% 50% 50% / 50%)', 'color(--hsl 180 0.5 0.5 / 0.5)'),
        ('color(--hsl none none none / none)', 'color(--hsl none none none / none)'),
        # Test range
        ('color(--hsl 0% 0% 0%)', 'color(--hsl 0 0 0)'),
        ('color(--hsl 100% 100% 100%)', 'color(--hsl 360 1 1)'),
        ('color(--hsl -100% -100% -100%)', 'color(--hsl -360 -1 -1)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        if color2 is None:
            with pytest.raises(ValueError):
                Color(color1)
        else:
            self.assertColorEqual(Color(color1).convert('hsl'), Color(color2), color=True)


class TestHSLSerialize(util.ColorAssertsPyTest):
    """Test HSL serialization."""

    COLORS = [
        # Test hex no options
        ('hsl(270 30% 75%)', {}, 'hsl(270 30% 75%)'),
        # Test alpha
        ('hsl(270 30% 75% / 0.5)', {}, 'hsl(270 30% 75% / 0.5)'),
        ('hsl(270 30% 75%)', {'alpha': True}, 'hsl(270 30% 75% / 1)'),
        ('hsl(270 30% 75% / 0.5)', {'alpha': False}, 'hsl(270 30% 75%)'),
        # Test percent
        ('hsl(270 30% 75% / 50%)', {'percent': False}, 'hsl(270 30 75 / 0.5)'),
        # Test None
        ('hsl(none 30% 75%)', {}, 'hsl(0 30% 75%)'),
        ('hsl(none 30% 75%)', {'none': True}, 'hsl(none 30% 75%)'),
        # Test fit
        ('hsl(20 150% 75%)', {}, 'hsl(19.619 100% 76.43%)'),
        ('hsl(20 150% 75%)', {'fit': False}, 'hsl(20 150% 75%)'),
        # Test legacy
        ('hsl(270 30% 75%)', {'comma': True}, 'hsl(270, 30%, 75%)'),
        # Test legacy alpha
        ('hsl(270 30% 75% / 0.5)', {'comma': True}, 'hsla(270, 30%, 75%, 0.5)'),
        ('hsl(270 30% 75%)', {'comma': True, 'alpha': True}, 'hsla(270, 30%, 75%, 1)'),
        ('hsl(270 30% 75% / 0.5)', {'comma': True, 'alpha': False}, 'hsl(270, 30%, 75%)'),
        # Test legacy percent
        ('hsl(270 30% 75% / 50%)', {'comma': True, 'percent': False}, 'hsla(270, 30%, 75%, 0.5)'),
        # Test legacy None
        ('hsl(none 30% 75%)', {'comma': True}, 'hsl(0, 30%, 75%)'),
        ('hsl(none 30% 75%)', {'comma': True, 'none': True}, 'hsl(0, 30%, 75%)'),
        # Test color
        ('hsl(none 30% 75% / 0.5)', {'color': True}, 'color(--hsl 0 0.3 0.75 / 0.5)'),
        ('hsl(none 30% 75%)', {'color': True, 'none': True}, 'color(--hsl none 0.3 0.75)'),
        ('hsl(0 30% 75%)', {'color': True, 'alpha': True}, 'color(--hsl 0 0.3 0.75 / 1)'),
        ('hsl(0 30% 75% / 0.5)', {'color': True, 'alpha': False}, 'color(--hsl 0 0.3 0.75)'),
        # Test Fit
        ('color(--hsl 0 -1.1 0.3)', {'color': True}, 'color(--hsl 180 1 0.3028)'),
        ('color(--hsl 0 -1.1 0.3)', {'color': True, 'fit': False}, 'color(--hsl 0 -1.1 0.3)'),
        # Test corner cases
        ('hsl(270deg-0.9%.9)', {'color': True, 'fit': False}, 'color(--hsl 270 -0.009 0.009)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestHSLProperties(util.ColorAsserts, unittest.TestCase):
    """Test HSL."""

    def test_names(self):
        """Test HSL-ish names."""

        self.assertEqual(Color('color(--hsl 120 50% 90% / 1)')._space.names(), ('h', 's', 'l'))

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c['hue'], 120)
        c['hue'] = 110
        self.assertEqual(c['hue'], 110)

    def test_saturation(self):
        """Test `saturation`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c['saturation'], 0.5)
        c['saturation'] = 0.6
        self.assertEqual(c['saturation'], 0.6)

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c['lightness'], 0.9)
        c['lightness'] = 0.8
        self.assertEqual(c['lightness'], 0.8)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--hsl 120 50% 90% / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('hsl', [NaN, 0.5, 1], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_none_input(self):
        """Test `none` null."""

        c = Color('hsl(none 0% 75% / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_sat(self):
        """Test minimum saturation."""

        c = Color('hsl(270 0% 75% / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_max_light(self):
        """Test maximum lightness."""

        c = Color('hsl(270 20% 100% / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_null_normalization_min_light(self):
        """Test minimum lightness."""

        c = Color('hsl(270 20% 0% / 1)').normalize()
        self.assertTrue(c.is_nan('hue'))

    def test_corner_case_null(self):
        """Test corner case that produces null."""

        c = Color('color(srgb -2 0 2)').convert('hsl')
        self.assertTrue(c.is_nan('hue'))


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('hsl', [270, 0.5, 0]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, 0.5, 1]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, 0, 0.5]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, 0.000001, 0.5]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, 0.5, 0.9999999]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, NaN, 0]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, NaN, 1]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, 0.0, NaN]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, 0.5, NaN]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, NaN, 0.5]).is_achromatic(), True)
        self.assertEqual(Color('hsl', [270, NaN, NaN]).is_achromatic(), True)



"""
Map our achromatic spline calculation against real achromatic response.

Report the delta between our spline and the real world. Also note the highest chroma climbs.
"""
import sys
import argparse
import os
import matplotlib.pyplot as plt

sys.path.insert(0, os.getcwd())

from coloraide.everything import ColorAll as Color  # noqa: E402
from coloraide.spaces.hct import HCT, xyz_to_hct, Achromatic  # noqa: E402


def main():
    """Main."""

    parser = argparse.ArgumentParser(
        prog='calc_hct_min_chroma.py',
        description='Calculate minimum chroma for achromatic colors in HCT and maps current spline against real values.'
    )
    # Flag arguments
    parser.add_argument(
        '--res', '-r', type=int, default=50000, help="Resolution to use when calculating range, default is 10000."
    )
    parser.add_argument(
        '--spline', '-S', type=str, default='catrom', help="Spline to use for approximation of achromatic line"
    )
    parser.add_argument(
        '--tuning', '-t', type=str, action='append',
        help="Spline tuning parameters: start:end:step:scale (int:int:int:float)"
    )
    parser.add_argument(
        '--dump', action='store_true', help="Dump calculated values."
    )
    args = parser.parse_args()

    return run(args.spline, args.tuning, args.res, args.dump)


def run(spline, tuning, res, dump):
    """Run."""

    tune = [[int(i) if e < 3 else float(i) for e, i in enumerate(x.split(':'))] for x in tuning]
    env = HCT.ENV
    test = Achromatic(spline=spline, env=env)
    test.calc_achromatic_response(tune, env=env)

    color = Color('srgb', [0, 0, 0])
    points1 = {}
    points2 = {}
    diff_over = 0
    diff_under = 0
    max_m = 0
    first = False

    for i in range(res + 1):
        div = res / 5
        v = i / div
        if v < 0.001:
            continue
        color.update('srgb', [v, v, v])
        xyz = color.convert('xyz-d65')
        h, c, t = xyz_to_hct(xyz[:-1], HCT.ENV)
        if not first:
            print('Starting T: ', t)
            print('Starting C: ', c)
            first = True

        if c > max_m:
            max_m = c

        domain = test.scale(t)
        calc = test.spline(domain)

        delta = calc[1] - c
        if delta >= 0 and delta > diff_over:
            diff_over = delta
        if delta < 0 and abs(delta) > diff_under:
            diff_under = abs(delta)

        points1[t] = (c, h)
        points2[calc[0]] = (calc[1], calc[2])

    print('Delta Over: ', diff_over)
    print('Delta Under: ', diff_under)
    print('Max M: ', max_m)
    print('Data Points: ', test.spline.length)

    t1 = []
    t2 = []
    m1 = []
    m2 = []
    h1 = []
    h2 = []
    for t in sorted(points1):
        t1.append(t)
        m1.append(points1[t][0])
        h1.append(points1[t][1])
    for t in sorted(points2):
        t2.append(t)
        m2.append(points2[t][0])
        h2.append(points2[t][1])

    figure = plt.figure()

    # Create axes
    ax = plt.axes(
        projection='3d',
        xlabel='C',
        ylabel='H',
        zlabel='T'
    )

    ax.set_aspect('auto')
    ax.set_title('HCT: Delta = {} - Max C = {}'.format(max(diff_over, diff_under), max_m))
    figure.add_axes(ax)

    # Print the calculated line against the real line
    plt.style.use('seaborn-v0_8-whitegrid')
    plt.plot(m1, h1, t1, '.', color='black')
    plt.plot(m2, h2, t2, '.', color='red', markersize=0.5)
    plt.show()

    if dump:
        print('===== Data =====')
        print(test.dump())

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
CAM 16 UCS.

https://observablehq.com/@jrus/cam16
https://arxiv.org/abs/1802.06067
https://doi.org/10.1002/col.22131
"""
from __future__ import annotations
import math
from . cam16 import CAM16
from ..types import Vector
from ..channels import Channel, FLG_MIRROR_PERCENT
from .. import algebra as alg

COEFFICENTS = {
    'lcd': (0.77, 0.007, 0.0053),
    'scd': (1.24, 0.007, 0.0363),
    'ucs': (1.00, 0.007, 0.0228)
}


def cam16_to_cam16_ucs(jab: Vector, model: str) -> Vector:
    """
    CAM16 (Jab) to CAM16 UCS (Jab).

    We can actually go between simply by removing the old colorfulness multiplier
    and then adding the new adjusted multiplier. Then we can just adjust lightness.
    """

    J, a, b = jab
    M = math.sqrt(a ** 2 + b ** 2)

    c1, c2 = COEFFICENTS[model][1:]

    if M != 0:
        a /= M
        b /= M
        M = math.log(1 + c2 * M) / c2
        a *= M
        b *= M

    return [
        (1 + 100 * c1) * J / (1 + c1 * J),
        a,
        b
    ]


def cam16_ucs_to_cam16(ucs: Vector, model: str) -> Vector:
    """
    CAM16 UCS (Jab) to CAM16 (Jab).

    We can actually go between simply by removing the old colorfulness multiplier
    and then adding the new adjusted multiplier. Then we can just adjust lightness.
    """

    J, a, b = ucs
    M = math.sqrt(a ** 2 + b ** 2)

    c1, c2 = COEFFICENTS[model][1:]

    if M != 0:
        a /= M
        b /= M
        M = (math.exp(M * c2) - 1) / c2
        a *= M
        b *= M

    return [
        J / (1 - c1 * (J - 100)),
        a,
        b
    ]


class CAM16UCS(CAM16):
    """CAM16 UCS (Jab) class."""

    BASE = "cam16"
    NAME = "cam16-ucs"
    SERIALIZE = ("--cam16-ucs",)
    MODEL = 'ucs'
    CHANNELS = (
        Channel("j", 0.0, 100.0, limit=(0.0, None)),
        Channel("a", -50.0, 50.0, flags=FLG_MIRROR_PERCENT),
        Channel("b", -50.0, 50.0, flags=FLG_MIRROR_PERCENT)
    )

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        j, a, b = cam16_ucs_to_cam16(coords, self.MODEL)
        m, h = alg.rect_to_polar(a, b)
        return coords[0] == 0.0 or self.ACHROMATIC.test(j, m, h)

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from CAM16."""

        return cam16_ucs_to_cam16(coords, self.MODEL)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to CAM16."""

        return cam16_to_cam16_ucs(coords, self.MODEL)


class CAM16LCD(CAM16UCS):
    """CAM16 LCD (Jab) class."""

    NAME = "cam16-lcd"
    SERIALIZE = ("--cam16-lcd",)
    MODEL = 'lcd'
    CHANNELS = (
        Channel("j", 0.0, 100.0, limit=(0.0, None)),
        Channel("a", -70.0, 70.0, flags=FLG_MIRROR_PERCENT),
        Channel("b", -70.0, 70.0, flags=FLG_MIRROR_PERCENT)
    )


class CAM16SCD(CAM16UCS):
    """CAM16 SCD (Jab) class."""

    NAME = "cam16-scd"
    SERIALIZE = ("--cam16-scd",)
    MODEL = 'scd'
    CHANNELS = (
        Channel("j", 0.0, 100.0, limit=(0.0, None)),
        Channel("a", -40.0, 40.0, flags=FLG_MIRROR_PERCENT),
        Channel("b", -40.0, 40.0, flags=FLG_MIRROR_PERCENT)
    )



"""
Ohno 2013 CCT calculations.

https://www.researchgate.net/publication/263373260_Practical_Use_and_Calculation_of_CCT_and_Duv
"""
from __future__ import annotations
import math
from . import planck
from .. import cat
from .. import cmfs
from .. import util
from .. import algebra as alg
from ..temperature import CCT
from ..types import Vector, VectorLike
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class BlackBodyCurve:
    """
    Setup a spline that represents the black body curve.

    Points between steps are approximated, but actual points can always be
    acquired via `exact`.

    For improved accuracy, we split spline data for low temps and high temps
    and assign the number of required data points accordingly.
    """

    def __init__(
        self,
        cmfs: dict[int, tuple[float, float, float]] = cmfs.CIE_1931_2DEG,
        white: VectorLike = cat.WHITES['2deg']['D65'],
        planck_step: int = 5,
        chromaticity: str = 'uv-1960'
    ) -> None:
        """Initialize."""

        keys = list(cmfs.keys())
        self.cmfs_start = min(keys)
        self.cmfs_end = max(keys)
        self.cmfs = cmfs
        self.white = util.xy_to_xyz(white)
        self.planck_step = planck_step
        self.to_uv = util.xy_to_uv_1960 if chromaticity == 'uv-1960' else util.xy_to_uv

        # Low temperature range
        start = 1000
        end = 20000
        step = 130
        inc = (end - start) / step
        count = step + 1
        points = []
        self.domain = []
        for r in range(count):
            k = r * inc + start
            u, v = self.to_uv(
                planck.temp_to_xy_planckian_locus(
                    k, self.cmfs, self.white, self.cmfs_start, self.cmfs_end, self.planck_step
                )
            )
            self.domain.append(k)
            points.append([u, v])
        self.spline = alg.interpolate(points, method='catrom')

        # High temperature range
        start = end
        end = 100000
        step = 220
        inc = (end - start) / step
        count = step + 1
        points = []
        self.domain2 = []
        for r in range(count):
            k = r * inc + start
            u, v = self.to_uv(
                planck.temp_to_xy_planckian_locus(
                    k, self.cmfs, self.white, self.cmfs_start, self.cmfs_end, self.planck_step
                )
            )
            self.domain2.append(k)
            points.append([u, v])
        self.spline2 = alg.interpolate(points, method='catrom')

    def scale(self, point: float, domain: list[float]) -> float:
        """Scale the temperature point to match the range 0 - 1."""

        # Extrapolation
        if point <= domain[0]:
            point = (point - domain[0]) / (domain[-1] - domain[0])
        elif point >= self.domain[-1]:
            point = 1.0 + (point - domain[-1]) / (domain[-1] - domain[0])

        # Interpolation
        else:
            a, b = domain[0], domain[len(domain) - 1]
            l = b - a
            point = ((point - a) / l) if l else 0.0
        return point

    def __call__(self, temp: float, exact: bool = False) -> Vector:
        """Get the uv for the given temp."""

        if exact:
            return self.to_uv(
                planck.temp_to_xy_planckian_locus(
                    temp, self.cmfs, self.white, self.cmfs_start, self.cmfs_end, self.planck_step
                )
            )
        else:
            if temp <= 20000:
                return self.spline(self.scale(temp, self.domain))
            return self.spline2(self.scale(temp, self.domain2))


class Ohno2013(CCT):
    """
    Calculate temperature for a given pair of uv coordinates.

    The Ohno approach requires a pre-generated table. The more data available, the more precise the values.
    Unfortunately, to span the entire range of 1000 - 100000 with fairly good accuracy, it requires keeping
    a very large table in memory.

    To avoid storing a large amount of data in memory, we can use multiple iterations and dynamically sample
    points on the locus, each iteration shrinking the bounds until we converge. Unfortunately, this is very
    slow, millisecond range.

    An alternative is to use the iterative approach, but generate a smaller subset of data and use a spline
    to approximate the points in between. Obviously, the points in between will not be as accurate, but the
    spline is used only as a way to approximate close to the temperature. Once we've sufficiently narrowed
    the range down to our best 3 temperature points, we can calculate those points with higher accuracy and
    proceed with the solvers. This actually allows us to use an even smaller amount of data than if we had
    used no spline and pre-calculated enough points for a similar accuracy. This is also much faster than
    dynamically calculating all the points.

    After navigating the table of data and determining a temperature that has the lowest delta distance, we can
    then use the triangular and parabolic solver. The triangular works best for values close to the locus (less
    than |0.002| Duv) and the parabolic solution works better for values with a higher Duv.

    For more precision, `exact` will avoid the approximation spline.

    https://www.researchgate.net/publication/263373260_Practical_Use_and_Calculation_of_CCT_and_Duv
    """

    NAME = 'ohno-2013'
    CHROMATICITY = 'uv-1960'

    def __init__(
        self,
        cmfs: dict[int, tuple[float, float, float]] = cmfs.CIE_1931_2DEG,
        white: VectorLike = cat.WHITES['2deg']['D65'],
        planck_step: int = 5
    ):
        """Initialize."""

        self.white = white
        self.blackbody = BlackBodyCurve(cmfs, white, planck_step, self.CHROMATICITY)

    def to_cct(
        self,
        color: Color,
        start: float = 1000,
        end: float = 100000,
        samples: int = 10,
        iterations: int = 6,
        exact: bool = False,
        **kwargs: Any
    ) -> Vector:
        """Calculate a color's CCT."""

        u, v = color.split_chromaticity(self.CHROMATICITY)[:-1]
        last = samples - 1
        index = 0
        table = []  # type: list[tuple[float, float, float, float]]

        # Each iteration we narrow the range until we are close enough
        for _ in range(iterations):
            table.clear()
            lowest = alg.INF
            index = 0

            # Generate the Planckian table while tracking lowest distance
            for j in range(samples):
                k = alg.lerp(start, end, j / last)
                u2, v2 = self.blackbody(k, exact=exact)
                di = math.sqrt((u2 - u) ** 2 + (v2 - v) ** 2)
                if di < lowest:
                    lowest = di
                    index = j
                table.append((k, u2, v2, di))

            # Set next iteration's range to include our best result +/-1
            # If our best result was on the edge, that edge remains the boundary
            start = table[index - 1][0] if index > 0 else table[index][0]
            end = table[index + 1][0] if index < last else table[index][0]

        # Select the closest 3 values. Get precise values instead of our
        # approximated spline value so we can get the most accurate result.
        ti = table[index][0]
        if not exact:
            ui, vi = self.blackbody(ti, exact=True)
            di = math.sqrt((ui - u) ** 2 + (vi - v) ** 2)
        else:
            di = table[index][-1]

        if index == 0 or not exact:
            tp = ti - 1e-4 if index == 0 else table[index - 1][0]
            up, vp = self.blackbody(tp, exact=True)
            dp = math.sqrt((up - u) ** 2 + (vp - v) ** 2)
        else:
            tp, up, vp, dp = table[index - 1]

        if index == last or not exact:
            tn = ti + 1e-4 if index == last else table[index + 1][0]
            un, vn = self.blackbody(tn, exact=True)
            dn = math.sqrt((un - u) ** 2 + (vn - v) ** 2)
        else:
            tn, un, vn, dn = table[index + 1]

        # Triangular solution
        l = math.sqrt((un - up) ** 2 + (vn - vp) ** 2)
        x = (dp ** 2 - dn ** 2 + l ** 2) / (2 * l)
        t = tp + (tn - tp) * (x / l)
        vtx = vp + (vn - vp) * (x / l)
        sign = math.copysign(1, v - vtx)
        duv = (dp ** 2 - x ** 2) ** (1 / 2) * sign

        # Parabolic solution
        if abs(duv) >= 0.002:
            x = (tn - ti) * (tp - tn) * (ti - tp)
            a = (
                tp * (dn - di) +
                ti * (dp - dn) +
                tn * (di - dp)
            ) * (x ** -1)
            b = -(
                (tp ** 2) * (dn - di) +
                (ti ** 2) * (dp - dn) +
                (tn ** 2) * (di - dp)
            ) * (x ** -1)
            c = -(
                (dp * ti * tn) * (tn - ti) +
                (di * tp * tn) * (tp - tn) +
                (dn * tp * ti) * (ti - tp)
            ) * (x ** -1)
            t = -b / (2 * a)
            duv = (a * (t ** 2) + b * t + c) * sign

        return [t, duv]

    def from_cct(
        self,
        color: type[Color],
        space: str,
        kelvin: float,
        duv: float,
        scale: bool,
        scale_space: str | None,
        **kwargs: Any
    ) -> Color:
        """Calculate a color that satisfies the CCT using Planck's law."""

        u0, v0 = self.blackbody(kelvin, exact=True)
        if duv:
            u1, v1 = self.blackbody(kelvin + 0.01, exact=True)
            du = u0 - u1
            dv = v0 - v1
            di = math.sqrt(du ** 2 + dv ** 2)
            if di:
                du /= di
                dv /= di
                u0 = u0 - duv * dv
                v0 = v0 + duv * du

        return color.chromaticity(space, [u0, v0, 1], self.CHROMATICITY, scale=scale, scale_space=scale_space)



"""Test utilities."""
import unittest
from coloraide import util, NaN


class TestUtil(unittest.TestCase):
    """Test utilities."""

    def test_compare_coords(self):
        """Test coordinate comparison."""

        self.assertTrue(util.cmp_coords([1, 2, 3], [1, 2, 3]))
        self.assertTrue(util.cmp_coords([1, NaN, 3], [1, NaN, 3]))
        self.assertFalse(util.cmp_coords([1, 3, 2], [1, 2, 3]))
        self.assertFalse(util.cmp_coords([1, 2, NaN], [1, NaN, 3]))
        self.assertFalse(util.cmp_coords([1, 2, 3, 4], [1, 2, 3]))
